import os

DATA_DIR = os.path.join(os.path.dirname(os.path.abspath(__file__)), "data")
IMGCALC_INDICES_DATA_DIR = os.path.join(DATA_DIR, "imagecalc", "calcindices")


def test_calc_ndvi(tmp_path):
    import rsgislib.imagecalc
    import rsgislib.imagecalc.calcindices

    input_img = os.path.join(DATA_DIR, "sen2_20210527_aber_subset.kea")
    ref_img = os.path.join(
        IMGCALC_INDICES_DATA_DIR, "sen2_20210527_aber_subset_ndvi_ref.kea"
    )

    output_img = os.path.join(tmp_path, "out_img.kea")
    rsgislib.imagecalc.calcindices.calc_ndvi(
        input_img, 3, 8, output_img, calc_stats=True, gdalformat="KEA"
    )

    img_eq, prop_match = rsgislib.imagecalc.are_img_bands_equal(
        output_img, 1, ref_img, 1
    )
    assert prop_match > 0.95


def test_calc_wbi(tmp_path):
    import rsgislib.imagecalc
    import rsgislib.imagecalc.calcindices

    input_img = os.path.join(DATA_DIR, "sen2_20210527_aber_subset.kea")
    ref_img = os.path.join(
        IMGCALC_INDICES_DATA_DIR, "sen2_20210527_aber_subset_wbi_ref.kea"
    )

    output_img = os.path.join(tmp_path, "out_img.kea")
    rsgislib.imagecalc.calcindices.calc_wbi(
        input_img, 1, 8, output_img, calc_stats=True, gdalformat="KEA"
    )

    img_eq, prop_match = rsgislib.imagecalc.are_img_bands_equal(
        output_img, 1, ref_img, 1
    )
    assert prop_match > 0.95


def test_calc_ndwi(tmp_path):
    import rsgislib.imagecalc
    import rsgislib.imagecalc.calcindices

    input_img = os.path.join(DATA_DIR, "sen2_20210527_aber_subset.kea")
    ref_img = os.path.join(
        IMGCALC_INDICES_DATA_DIR, "sen2_20210527_aber_subset_ndwi_ref.kea"
    )

    output_img = os.path.join(tmp_path, "out_img.kea")
    rsgislib.imagecalc.calcindices.calc_ndwi(
        input_img, 8, 9, output_img, calc_stats=True, gdalformat="KEA"
    )

    img_eq, prop_match = rsgislib.imagecalc.are_img_bands_equal(
        output_img, 1, ref_img, 1
    )
    assert prop_match > 0.95


def test_calc_gndwi(tmp_path):
    import rsgislib.imagecalc
    import rsgislib.imagecalc.calcindices

    input_img = os.path.join(DATA_DIR, "sen2_20210527_aber_subset.kea")
    ref_img = os.path.join(
        IMGCALC_INDICES_DATA_DIR, "sen2_20210527_aber_subset_gndwi_ref.kea"
    )

    output_img = os.path.join(tmp_path, "out_img.kea")
    rsgislib.imagecalc.calcindices.calc_gndwi(
        input_img, 2, 8, output_img, calc_stats=True, gdalformat="KEA"
    )

    img_eq, prop_match = rsgislib.imagecalc.are_img_bands_equal(
        output_img, 1, ref_img, 1
    )
    assert prop_match > 0.95


def test_calc_gmndwi(tmp_path):
    import rsgislib.imagecalc
    import rsgislib.imagecalc.calcindices

    input_img = os.path.join(DATA_DIR, "sen2_20210527_aber_subset.kea")
    ref_img = os.path.join(
        IMGCALC_INDICES_DATA_DIR, "sen2_20210527_aber_subset_gmndwi_ref.kea"
    )

    output_img = os.path.join(tmp_path, "out_img.kea")
    rsgislib.imagecalc.calcindices.calc_gmndwi(
        input_img, 2, 9, output_img, calc_stats=True, gdalformat="KEA"
    )

    img_eq, prop_match = rsgislib.imagecalc.are_img_bands_equal(
        output_img, 1, ref_img, 1
    )
    assert prop_match > 0.95


def test_calc_whiteness(tmp_path):
    import rsgislib.imagecalc
    import rsgislib.imagecalc.calcindices

    input_img = os.path.join(DATA_DIR, "sen2_20210527_aber_subset.kea")
    ref_img = os.path.join(
        IMGCALC_INDICES_DATA_DIR, "sen2_20210527_aber_subset_whiteness_ref.kea"
    )

    output_img = os.path.join(tmp_path, "out_img.kea")
    rsgislib.imagecalc.calcindices.calc_whiteness(
        input_img, 1, 2, 3, output_img, calc_stats=True, gdalformat="KEA"
    )

    img_eq, prop_match = rsgislib.imagecalc.are_img_bands_equal(
        output_img, 1, ref_img, 1
    )
    assert prop_match > 0.95


def test_calc_brightness(tmp_path):
    import rsgislib.imagecalc
    import rsgislib.imagecalc.calcindices

    input_img = os.path.join(DATA_DIR, "sen2_20210527_aber_subset.kea")
    ref_img = os.path.join(
        IMGCALC_INDICES_DATA_DIR, "sen2_20210527_aber_subset_brightness_ref.kea"
    )

    output_img = os.path.join(tmp_path, "out_img.kea")
    rsgislib.imagecalc.calcindices.calc_brightness(
        input_img,
        1,
        2,
        3,
        output_img,
        calc_stats=True,
        gdalformat="KEA",
        scale_factor=1000,
    )

    img_eq, prop_match = rsgislib.imagecalc.are_img_bands_equal(
        output_img, 1, ref_img, 1
    )
    assert prop_match > 0.95


def test_calc_brightness_scaled(tmp_path):
    import rsgislib.imagecalc
    import rsgislib.imagecalc.calcindices

    input_img = os.path.join(DATA_DIR, "sen2_20210527_aber_subset.kea")
    ref_img = os.path.join(
        IMGCALC_INDICES_DATA_DIR, "sen2_20210527_aber_subset_brightness_scaled_ref.kea"
    )

    output_img = os.path.join(tmp_path, "out_img.kea")
    rsgislib.imagecalc.calcindices.calc_brightness_scaled(
        input_img,
        1,
        2,
        3,
        output_img,
        calc_stats=True,
        gdalformat="KEA",
        scale_factor=1000,
    )

    img_eq, prop_match = rsgislib.imagecalc.are_img_bands_equal(
        output_img, 1, ref_img, 1
    )
    assert prop_match > 0.95


def test_calc_ctvi(tmp_path):
    import rsgislib.imagecalc
    import rsgislib.imagecalc.calcindices

    input_img = os.path.join(DATA_DIR, "sen2_20210527_aber_subset.kea")
    ref_img = os.path.join(
        IMGCALC_INDICES_DATA_DIR, "sen2_20210527_aber_subset_ctvi_ref.kea"
    )

    output_img = os.path.join(tmp_path, "out_img.kea")
    rsgislib.imagecalc.calcindices.calc_ctvi(
        input_img, 1, 8, output_img, calc_stats=True, gdalformat="KEA"
    )

    img_eq, prop_match = rsgislib.imagecalc.are_img_bands_equal(
        output_img, 1, ref_img, 1
    )
    assert prop_match > 0.95


def test_calc_ndsi(tmp_path):
    import rsgislib.imagecalc
    import rsgislib.imagecalc.calcindices

    input_img = os.path.join(DATA_DIR, "sen2_20210527_aber_subset.kea")
    ref_img = os.path.join(
        IMGCALC_INDICES_DATA_DIR, "sen2_20210527_aber_subset_ndsi_ref.kea"
    )

    output_img = os.path.join(tmp_path, "out_img.kea")
    rsgislib.imagecalc.calcindices.calc_ndsi(
        input_img, 2, 9, output_img, calc_stats=True, gdalformat="KEA"
    )

    img_eq, prop_match = rsgislib.imagecalc.are_img_bands_equal(
        output_img, 1, ref_img, 1
    )
    assert prop_match > 0.95


def test_calc_nbr(tmp_path):
    import rsgislib.imagecalc
    import rsgislib.imagecalc.calcindices

    input_img = os.path.join(DATA_DIR, "sen2_20210527_aber_subset.kea")
    ref_img = os.path.join(
        IMGCALC_INDICES_DATA_DIR, "sen2_20210527_aber_subset_nbr_ref.kea"
    )

    output_img = os.path.join(tmp_path, "out_img.kea")
    rsgislib.imagecalc.calcindices.calc_nbr(
        input_img, 8, 10, output_img, calc_stats=True, gdalformat="KEA"
    )

    img_eq, prop_match = rsgislib.imagecalc.are_img_bands_equal(
        output_img, 1, ref_img, 1
    )
    assert prop_match > 0.95


def test_calc_bai(tmp_path):
    import rsgislib.imagecalc
    import rsgislib.imagecalc.calcindices

    input_img = os.path.join(DATA_DIR, "sen2_20210527_aber_subset.kea")
    ref_img = os.path.join(
        IMGCALC_INDICES_DATA_DIR, "sen2_20210527_aber_subset_bai_ref.kea"
    )

    output_img = os.path.join(tmp_path, "out_img.kea")
    rsgislib.imagecalc.calcindices.calc_bai(
        input_img,
        8,
        3,
        output_img,
        calc_stats=True,
        gdalformat="KEA",
        scale_factor=1000,
    )

    img_eq, prop_match = rsgislib.imagecalc.are_img_bands_equal(
        output_img, 1, ref_img, 1
    )
    assert prop_match > 0.95


def test_calc_mvi(tmp_path):
    import rsgislib.imagecalc
    import rsgislib.imagecalc.calcindices

    input_img = os.path.join(DATA_DIR, "sen2_20210527_aber_subset.kea")
    ref_img = os.path.join(
        IMGCALC_INDICES_DATA_DIR, "sen2_20210527_aber_subset_mvi_ref.kea"
    )

    output_img = os.path.join(tmp_path, "out_img.kea")
    rsgislib.imagecalc.calcindices.calc_mvi(
        input_img, 2, 8, 9, output_img, calc_stats=True, gdalformat="KEA"
    )

    img_eq, prop_match = rsgislib.imagecalc.are_img_bands_equal(
        output_img, 1, ref_img, 1
    )
    assert prop_match > 0.95



#!/usr/bin/env python
"""
The tools.validate_datesets module contains functions for checking that datasets are valid.
"""

import os

from osgeo import gdal
from osgeo import osr
import numpy

import rsgislib
import rsgislib.tools.filetools


def check_gdal_image_file(
    input_img: str,
    check_bands: bool = True,
    n_bands: int = 0,
    chk_proj: bool = False,
    epsg_code: int = 0,
    read_img: bool = False,
    smpl_n_pxls: int = 10,
    calc_chk_sum: bool = False,
    max_file_size: int = None,
):
    """
    A function which checks a GDAL compatible image file and returns an error
    message if appropriate.

    :param input_img: the file path to the gdal image file.
    :param check_bands: boolean specifying whether individual image bands should be
                        opened and checked (Default: True)
    :param n_bands: int specifying the number of expected image bands. Ignored if 0;
                    Default is 0.
    :param chk_proj: boolean specifying whether to check that the projection has
                     been defined.
    :param epsg_code: int for the EPSG code for the projection. Error raised if image
                      is not that projection.
    :param read_img: boolean specifying whether to try reading some image pixel values
                     from the image. This option will read npxls (e.g., 10) random
                     image pixel values from a randomly selected band.
    :param smpl_n_pxls: The number of pixel values to be randomly selected (default =
                        10). More values = longer runtime.
    :param calc_chk_sum: boolean specifying whether a checksum should be calculated
                         for each band to check validity
    :param max_file_size: int specifying the maximum file size for the input file.
                          If None then ignored.
    :return: boolean (True: file ok; False: Error found), string (error message if
                      required otherwise empty string)

    """
    file_ok = True
    err_str = ""
    if os.path.exists(input_img):
        if max_file_size is not None:
            file_size_bytes = rsgislib.tools.filetools.get_file_size(input_img)
            if file_size_bytes == 0:
                file_ok = False
                err_str = "File size is zero."
            elif file_size_bytes > max_file_size:
                file_ok = False
                err_str = "File size is greater than max. threshold ({}).".format(
                    file_size_bytes
                )
        if file_ok:
            err = rsgislib.RSGISGDALErrorHandler()
            err_handler = err.handler
            gdal.PushErrorHandler(err_handler)
            gdal.UseExceptions()
            try:
                if os.path.splitext(input_img)[1].lower() == ".kea":
                    file_ok = check_hdf5_file(input_img)
                    if not file_ok:
                        err_str = "Error with KEA/HDF5 file."
                if file_ok:
                    raster_ds = gdal.Open(input_img, gdal.GA_ReadOnly)
                    if raster_ds is None:
                        file_ok = False
                        err_str = "GDAL could not open the dataset, returned None."

                    if file_ok and (n_bands > 0):
                        n_img_bands = raster_ds.RasterCount
                        if n_img_bands != n_bands:
                            file_ok = False
                            err_str = (
                                "Image should have {} image bands "
                                "but {} found.".format(n_bands, n_img_bands)
                            )

                    if file_ok and check_bands:
                        n_img_bands = raster_ds.RasterCount
                        if n_img_bands < 1:
                            file_ok = False
                            err_str = "Image says it does not have any image bands."
                        else:
                            for n in range(n_img_bands):
                                band = n + 1
                                img_band = raster_ds.GetRasterBand(band)
                                if img_band is None:
                                    file_ok = False
                                    err_str = (
                                        "GDAL could not open band {} in "
                                        "the dataset, returned None.".format(band)
                                    )
                                    break

                    if file_ok and chk_proj:
                        proj_obj = raster_ds.GetProjection()
                        if proj_obj is None:
                            file_ok = False
                            err_str = "Image projection is None."
                        elif proj_obj == "":
                            file_ok = False
                            err_str = "Image projection is empty."

                        if file_ok and (epsg_code > 0):
                            spat_ref = osr.SpatialReference()
                            spat_ref.ImportFromWkt(proj_obj)
                            spat_ref.AutoIdentifyEPSG()
                            img_epsg_code = spat_ref.GetAuthorityCode(None)
                            if img_epsg_code is None:
                                file_ok = False
                                err_str = "Image projection returned a None EPSG code."
                            elif int(img_epsg_code) != int(epsg_code):
                                file_ok = False
                                err_str = (
                                    "Image EPSG ({}) does not match "
                                    "that specified ({})".format(
                                        img_epsg_code, epsg_code
                                    )
                                )

                    if file_ok and read_img:
                        n_img_bands = raster_ds.RasterCount
                        xSize = raster_ds.RasterXSize
                        ySize = raster_ds.RasterYSize

                        if n_img_bands == 1:
                            band = 1
                        else:
                            band = int(
                                numpy.random.randint(1, high=n_img_bands, size=1)
                            )

                        img_band = raster_ds.GetRasterBand(band)
                        x_pxls = numpy.random.choice(xSize, smpl_n_pxls)
                        y_pxls = numpy.random.choice(ySize, smpl_n_pxls)
                        for i in range(smpl_n_pxls):
                            img_data = img_band.ReadRaster(
                                xoff=int(x_pxls[i]),
                                yoff=int(y_pxls[i]),
                                xsize=1,
                                ysize=1,
                                buf_xsize=1,
                                buf_ysize=1,
                                buf_type=gdal.GDT_Float32,
                            )

                    if file_ok and calc_chk_sum:
                        n_img_bands = raster_ds.RasterCount
                        for n in range(n_img_bands):
                            raster_ds.GetRasterBand(n + 1).Checksum()

                    raster_ds = None
            except Exception as e:
                file_ok = False
                err_str = str(e)
            else:
                if err.err_level >= gdal.CE_Warning:
                    file_ok = False
                    err_str = str(err.err_msg)
            finally:
                gdal.PopErrorHandler()
    else:
        file_ok = False
        err_str = "File does not exist."
    return file_ok, err_str


def _run_check_gdal_image_file(img_params: list):
    """
    Internal function called by run_check_gdal_image_file - not intended to be called
    by end user.

    :param img_params: list of function inputs.
    :return: boolean whether the specified file is OK or not.

    """
    input_img = img_params[0]
    check_bands = img_params[1]
    n_bands = img_params[2]
    chk_proj = img_params[3]
    epsg_code = img_params[4]
    read_img = img_params[5]
    smpl_n_pxls = img_params[6]
    calc_chk_sum = img_params[7]
    max_file_size = img_params[8]
    rm_err = img_params[9]
    print_err = img_params[10]

    try:
        file_ok, err_str = check_gdal_image_file(
            input_img,
            check_bands=check_bands,
            n_bands=n_bands,
            chk_proj=chk_proj,
            epsg_code=epsg_code,
            read_img=read_img,
            smpl_n_pxls=smpl_n_pxls,
            calc_chk_sum=calc_chk_sum,
            max_file_size=max_file_size,
        )
        if print_err and (not file_ok):
            print("Error: '{}'".format(err_str))
        if not file_ok:
            if rm_err:
                rsgislib.tools.filetools.delete_file_silent(input_img)
                print("Removed {}".format(input_img))
            else:
                print("rm {}".format(input_img))
    except Exception as e:
        if print_err:
            print("Error: '{}'".format(e))
        if rm_err:
            rsgislib.tools.filetools.delete_file_silent(input_img)
            print("Removed {}".format(input_img))
        else:
            print("rm {}".format(input_img))
        file_ok = False

    return file_ok


def run_check_gdal_image_file(
    input_img: str,
    check_bands: bool = True,
    n_bands: int = 0,
    chk_proj: bool = False,
    epsg_code: int = 0,
    read_img: bool = False,
    smpl_n_pxls: int = 10,
    calc_chk_sum: bool = False,
    max_file_size: int = None,
    rm_err: bool = False,
    print_err: bool = True,
    timeout: int = 4,
):
    """
    A function which checks a GDAL compatible image file using the check_gdal_image_file
    function where a mutliprocessing object is used to catch errors which can crash
    Python and still continue without crashing the Python environment.

    You probably want to call this function rather than calling check_gdal_image_file
    directly.

    :param input_img: the file path to the gdal image file.
    :param check_bands: boolean specifying whether individual image bands should be
                        opened and checked (Default: True)
    :param n_bands: int specifying the number of expected image bands. Ignored if 0;
                    Default is 0.
    :param chk_proj: boolean specifying whether to check that the projection has
                     been defined.
    :param epsg_code: int for the EPSG code for the projection. Error raised if image
                      is not that projection.
    :param read_img: boolean specifying whether to try reading some image pixel values
                     from the image. This option will read npxls (e.g., 10) random
                     image pixel values from a randomly selected band.
    :param smpl_n_pxls: The number of pixel values to be randomly selected (default =
                        10). More values = longer runtime.
    :param calc_chk_sum: boolean specifying whether a checksum should be calculated
                         for each band to check validity
    :param max_file_size: int specifying the maximum file size for the input file.
                          If None then ignored.
    :param rm_err: boolean specifying whether to delete the file if an error is found
    :param print_err: print any errors associated with the file to the console
    :param timeout: a timeout in seconds (Default = 4) for the tests to be undertaken.
    :return: boolean whether the file is OK (i.e., passed tests) or not.

    """
    from multiprocessing import Pool

    processes_pool = Pool(1)
    params = [
        input_img,
        check_bands,
        n_bands,
        chk_proj,
        epsg_code,
        read_img,
        smpl_n_pxls,
        calc_chk_sum,
        max_file_size,
        rm_err,
        print_err,
    ]
    try:
        result = processes_pool.apply_async(_run_check_gdal_image_file, args=[params])
        file_ok = result.get(timeout=timeout)
    except Exception as e:
        file_ok = False
        if rm_err:
            rsgislib.tools.filetools.delete_file_silent(input_img)
            print("Removed {}".format(input_img))
        else:
            print("rm {}".format(input_img))
    processes_pool.close()
    processes_pool.join()

    return file_ok


def run_check_gdal_image_files(
    input_imgs: list,
    check_bands: bool = True,
    n_bands: int = 0,
    chk_proj: bool = False,
    epsg_code: int = 0,
    read_img: bool = False,
    smpl_n_pxls: int = 10,
    calc_chk_sum: bool = False,
    max_file_size: int = None,
    rm_err: bool = False,
    print_err: bool = True,
    print_file_names: bool = False,
    timeout: int = 4,
):
    """
    A function which checks a list of GDAL compatible image files using the
    check_gdal_image_file function where a mutliprocessing object is used to
    catch errors which can crash Python and still continue without crashing the
    Python environment.

    You probably want to call this function rather than calling check_gdal_image_file
    directly.

    :param input_imgs: a list of input images.
    :param check_bands: boolean specifying whether individual image bands should be
                        opened and checked (Default: True)
    :param n_bands: int specifying the number of expected image bands. Ignored if 0;
                    Default is 0.
    :param chk_proj: boolean specifying whether to check that the projection has
                     been defined.
    :param epsg_code: int for the EPSG code for the projection. Error raised if image
                      is not that projection.
    :param read_img: boolean specifying whether to try reading some image pixel values
                     from the image. This option will read npxls (e.g., 10) random
                     image pixel values from a randomly selected band.
    :param smpl_n_pxls: The number of pixel values to be randomly selected (default =
                        10). More values = longer runtime.
    :param calc_chk_sum: boolean specifying whether a checksum should be calculated
                         for each band to check validity
    :param max_file_size: int specifying the maximum file size for the input file.
                          If None then ignored.
    :param rm_err: boolean specifying whether to delete the file if an error is found
    :param print_err: print any errors associated with the file to the console
    :param print_file_names: print the names of the file before they are tested.
    :param timeout: a timeout in seconds (Default = 4) for the tests to be undertaken.
    :return: boolean whether all the files are OK (i.e., passed tests) or not.

    """
    from multiprocessing import Pool

    files_ok = True
    try:
        processes_pool = Pool(1)
        for input_img in input_imgs:
            if print_file_names:
                print(input_img)
            params = [
                input_img,
                check_bands,
                n_bands,
                chk_proj,
                epsg_code,
                read_img,
                smpl_n_pxls,
                calc_chk_sum,
                max_file_size,
                rm_err,
                print_err,
            ]
            try:
                result = processes_pool.apply_async(
                    _run_check_gdal_image_file, args=[params]
                )
                file_ok = result.get(timeout=timeout)
                if not file_ok:
                    files_ok = False
            except Exception as e:
                files_ok = False
                if rm_err:
                    rsgislib.tools.filetools.delete_file_silent(input_img)
                    print("Removed {}".format(input_img))
                else:
                    print("rm {}".format(input_img))
        processes_pool.close()
        processes_pool.join()
    except Exception as inst:
        print("Finished with pool")

    return files_ok


def check_gdal_vector_file(
    vec_file: str, chk_proj: bool = True, epsg_code: int = 0, max_file_size: int = None
):
    """
    A function which checks a GDAL compatible vector file and returns an
    error message if appropriate.

    :param vec_file: the file path to the gdal vector file.
    :param chk_proj: boolean specifying whether to check that the projection has
                     been defined.
    :param epsg_code: int for the EPSG code for the projection. Error raised if image
                      is not that projection.
    :param max_file_size: int specifying the maximum file size for the input file.
                          If None then ignored.
    :return: boolean (True: file OK; False: Error found), string (error message
             if required otherwise empty string)

    """
    file_ok = True
    err_str = ""
    if os.path.exists(vec_file):
        if max_file_size is not None:
            file_size_bytes = rsgislib.tools.filetools.get_file_size(vec_file)
            if file_size_bytes == 0:
                file_ok = False
                err_str = "File size is zero."
            elif file_size_bytes > max_file_size:
                file_ok = False
                err_str = "File size is greater than max. threshold ({}).".format(
                    file_size_bytes
                )
        if file_ok:
            err = rsgislib.RSGISGDALErrorHandler()
            err_handler = err.handler
            gdal.PushErrorHandler(err_handler)
            gdal.UseExceptions()

            try:
                vec_ds = gdal.OpenEx(vec_file, gdal.OF_VECTOR)
                if vec_ds is None:
                    file_ok = False
                    err_str = "GDAL could not open the data source, returned None."
                else:
                    for lyr_idx in range(vec_ds.GetLayerCount()):
                        vec_lyr = vec_ds.GetLayerByIndex(lyr_idx)
                        if vec_lyr is None:
                            file_ok = False
                            err_str = "GDAL could not open all the vector layers."
                            break
                    if file_ok and chk_proj:
                        vec_lyr = vec_ds.GetLayer()
                        if vec_lyr is None:
                            raise rsgislib.RSGISPyException(
                                "Something has gone wrong "
                                "checking projection - layer not present"
                            )
                        vec_lyr_spt_ref = vec_lyr.GetSpatialRef()
                        if vec_lyr_spt_ref is None:
                            file_ok = False
                            err_str = "Vector projection is None."
                        if file_ok:
                            spt_ref_wkt = vec_lyr_spt_ref.ExportToWkt()
                            if spt_ref_wkt is None:
                                file_ok = False
                                err_str = "Vector projection WKT is None."
                            elif spt_ref_wkt == "":
                                file_ok = False
                                err_str = "Vector projection is empty."

                            if file_ok and (epsg_code > 0):
                                vec_lyr_spt_ref.AutoIdentifyEPSG()
                                vec_epsg_code = vec_lyr_spt_ref.GetAuthorityCode(None)
                                if vec_epsg_code is None:
                                    file_ok = False
                                    err_str = (
                                        "Vector projection returned a None EPSG code."
                                    )
                                elif int(vec_epsg_code) != int(epsg_code):
                                    file_ok = False
                                    err_str = (
                                        "Vector EPSG ({}) does not match "
                                        "that specified ({})".format(
                                            vec_epsg_code, epsg_code
                                        )
                                    )

                vec_ds = None
            except Exception as e:
                file_ok = False
                err_str = str(e)
            else:
                if err.err_level >= gdal.CE_Warning:
                    file_ok = False
                    err_str = str(err.err_msg)
            finally:
                gdal.PopErrorHandler()
    else:
        file_ok = False
        err_str = "File does not exist."
    return file_ok, err_str


def _run_check_gdal_vector_file(file_params: list):
    """
    Internal function called by check_gdal_vector_file - not intended to be called
    by end user.

    :param file_params: list of function inputs.
    :return: boolean whether the specified file is OK or not.

    """
    vec_file = file_params[0]
    chk_proj = file_params[1]
    epsg_code = file_params[2]
    max_file_size = file_params[3]
    rm_err = file_params[4]
    print_err = file_params[5]
    multi_file = file_params[6]

    try:
        file_ok, err_str = check_gdal_vector_file(
            vec_file,
            chk_proj=chk_proj,
            epsg_code=epsg_code,
            max_file_size=max_file_size,
        )
        if print_err and (not file_ok):
            print("Error: '{}'".format(err_str))
        if not file_ok:
            if rm_err:
                if multi_file:
                    rsgislib.tools.filetools.delete_file_with_basename(vec_file)
                else:
                    rsgislib.tools.filetools.delete_file_silent(vec_file)
                    print("Removed {}".format(vec_file))
            else:
                print("rm {}".format(vec_file))
    except Exception as e:
        if print_err:
            print("Error: '{}'".format(e))
        if rm_err:
            if multi_file:
                rsgislib.tools.filetools.delete_file_with_basename(vec_file)
            else:
                rsgislib.tools.filetools.delete_file_silent(vec_file)
                print("Removed {}".format(vec_file))
        else:
            print("rm {}".format(vec_file))
        file_ok = False

    return file_ok


def run_check_gdal_vector_file(
    vec_file: str,
    chk_proj: bool = True,
    epsg_code: int = 0,
    max_file_size: int = None,
    rm_err: bool = False,
    print_err: bool = True,
    multi_file: bool = False,
    timeout: int = 4,
):
    """
    A function which checks a GDAL compatible vector file using the
    check_gdal_vector_file function where a mutliprocessing object is used to
    catch errors which can crash Python and still continue without crashing
    the Python environment.

    You probably want to call this function rather than calling check_gdal_vector_file
    directly.

    :param vec_file: the file path to the gdal vector file.
    :param chk_proj: boolean specifying whether to check that the projection has
                     been defined.
    :param epsg_code: int for the EPSG code for the projection. Error raised if image
                      is not that projection.
    :param max_file_size: int specifying the maximum file size for the input file.
                          If None then ignored.
    :param rm_err: boolean specifying whether to delete the file if an error is found
    :param print_err: print any errors associated with the file to the console
    :param multi_file: if True (Default: False) then remove files with the same
                       basename. Useful for ESRI Shapefiles which are made up of
                       multiple files.
    :param timeout: a timeout in seconds (Default = 4) for the tests to be undertaken.
    :return: boolean specifying whether the file is OK (i.e., tests passed) or not.

    """
    from multiprocessing import Pool

    processes_pool = Pool(1)
    params = [
        vec_file,
        chk_proj,
        epsg_code,
        max_file_size,
        rm_err,
        print_err,
        multi_file,
    ]
    try:
        result = processes_pool.apply_async(_run_check_gdal_vector_file, args=[params])
        file_ok = result.get(timeout=timeout)
    except Exception as e:
        file_ok = False
        if rm_err:
            if multi_file:
                rsgislib.tools.filetools.delete_file_with_basename(vec_file)
            else:
                rsgislib.tools.filetools.delete_file_silent(vec_file)
                print("Removed {}".format(vec_file))
        else:
            print("rm {}".format(vec_file))
    processes_pool.close()
    processes_pool.join()

    return file_ok


def run_check_gdal_vector_files(
    vec_files: list,
    chk_proj: bool = True,
    epsg_code: int = 0,
    max_file_size: int = None,
    rm_err: bool = False,
    print_err: bool = True,
    multi_file: bool = False,
    print_file_names: bool = False,
    timeout: int = 4,
):
    """
    A function which checks a list of GDAL compatible vector files using the
    check_gdal_vector_file function where a mutliprocessing object is used to
    catch errors which can crash Python and still continue without crashing
    the Python environment.

    You probably want to call this function rather than calling check_gdal_vector_file
    directly.

    :param vec_files: list of input file paths.
    :param chk_proj: boolean specifying whether to check that the projection has
                     been defined.
    :param epsg_code: int for the EPSG code for the projection. Error raised if image
                      is not that projection.
    :param max_file_size: int specifying the maximum file size for the input file.
                          If None then ignored.
    :param rm_err: boolean specifying whether to delete the file if an error is found
    :param print_err: print any errors associated with the file to the console
    :param multi_file: if True (Default: False) then remove files with the same
                       basename. Useful for ESRI Shapefiles which are made up of
                       multiple files.
    :param print_file_names: print the names of the file before they are tested.
    :param timeout: a timeout in seconds (Default = 4) for the tests to be undertaken.
    :return: boolean specifying whether all the files are OK
             (i.e., tests passed) or not.

    """
    from multiprocessing import Pool

    files_ok = True
    try:
        processes_pool = Pool(1)
        for vec_file in vec_files:
            if print_file_names:
                print(vec_file)
            params = [
                vec_file,
                chk_proj,
                epsg_code,
                max_file_size,
                rm_err,
                print_err,
                multi_file,
            ]
            try:
                result = processes_pool.apply_async(
                    _run_check_gdal_vector_file, args=[params]
                )
                file_ok = result.get(timeout=timeout)
                if not file_ok:
                    files_ok = False
            except Exception as e:
                files_ok = False
                if rm_err:
                    if multi_file:
                        rsgislib.tools.filetools.delete_file_with_basename(vec_file)
                    else:
                        rsgislib.tools.filetools.delete_file_silent(vec_file)
                        print("Removed {}".format(vec_file))
                else:
                    print("rm {}".format(vec_file))
        processes_pool.close()
        processes_pool.join()
    except Exception as inst:
        print("Finished with pool")

    return files_ok


def check_hdf5_file(input_file: str):
    """
    A function which checks whether a HDF5 file is valid.

    :param input_file: the file path to the input file.
    :return: a boolean - True file is valid. False file is not valid.

    """
    import h5py

    def _check_h5_var(h5_obj):
        lcl_ok = True
        lcl_err = ""
        try:
            if isinstance(h5_obj, h5py.Dataset):
                lcl_ok = True
            elif isinstance(h5_obj, h5py.Group):
                for var in h5_obj.keys():
                    lcl_ok, lcl_err = _check_h5_var(h5_obj[var])
                    if not lcl_ok:
                        break
        except RuntimeError as e:
            lcl_ok = False
            lcl_err = str(e)
        except Exception as e:
            lcl_ok = False
            lcl_err = str(e)
        return lcl_ok, lcl_err

    glb_ok = True
    err_str = ""
    if not os.path.exists(input_file):
        glb_ok = False
    else:
        try:
            fH5 = h5py.File(input_file, "r")
            if fH5 is None:
                glb_ok = False
                err_str = "Could not open HDF5 file"
            else:
                for var in fH5.keys():
                    glb_ok, err_str = _check_h5_var(fH5[var])
                    if not glb_ok:
                        break
        except RuntimeError as e:
            glb_ok = False
            err_str = str(e)
        except Exception as e:
            glb_ok = False
            err_str = str(e)
    return glb_ok, err_str


def _run_check_hdf5_file(file_params: list):
    """
    Internal function called by check_hdf5_file - not intended to be called
    by end user.

    :param file_params: list of function inputs.
    :return: boolean whether the specified file is OK or not.

    """
    input_file = file_params[0]
    rm_err = file_params[1]
    print_err = file_params[2]

    try:
        file_ok, err_str = check_hdf5_file(input_file)
        if print_err and (not file_ok):
            print("Error: '{}'".format(err_str))
        if not file_ok:
            if rm_err:
                rsgislib.tools.filetools.delete_file_silent(input_file)
                print("Removed {}".format(input_file))
            else:
                print("rm {}".format(input_file))
    except Exception as e:
        if print_err:
            print("Error: '{}'".format(e))
        if rm_err:
            rsgislib.tools.filetools.delete_file_silent(input_file)
            print("Removed {}".format(input_file))
        else:
            print("rm {}".format(input_file))
        file_ok = False

    return file_ok


def run_check_hdf5_file(
    input_file: str, rm_err: bool = False, print_err: bool = True, timeout: int = 4
):
    """
    A function which checks a HDF5 file using the check_hdf5_file function
    where a mutliprocessing object is used to catch errors which can crash Python
    and still continue without crashing the Python environment.

    You probably want to call this function rather than calling check_hdf5_file
    directly.

    :param input_file: the file path to the HDF5 file.
    :param rm_err: boolean specifying whether to delete the file if an error is found
    :param print_err: print any errors associated with the file to the console
    :param timeout: a timeout in seconds (Default = 4) for the tests to be undertaken.
    :return: boolean specifying whether the file is OK (i.e., tests passed) or not.

    """
    from multiprocessing import Pool

    processes_pool = Pool(1)
    params = [input_file, rm_err, print_err]
    try:
        result = processes_pool.apply_async(_run_check_hdf5_file, args=[params])
        file_ok = result.get(timeout=timeout)
    except Exception as e:
        file_ok = False
        if rm_err:
            rsgislib.tools.filetools.delete_file_silent(input_file)
            print("Removed {}".format(input_file))
        else:
            print("rm {}".format(input_file))
    processes_pool.close()
    processes_pool.join()

    return file_ok


def run_check_hdf5_files(
    input_files: list,
    rm_err: bool = False,
    print_err: bool = True,
    print_file_names: bool = False,
    timeout: int = 4,
):
    """
    A function which checks a list of HDF5 files using the
    check_hdf5_file function where a mutliprocessing object is used to
    catch errors which can crash Python and still continue without crashing
    the Python environment.

    You probably want to call this function rather than calling check_hdf5_file
    directly.

    :param input_files: a list of input HDF5 file paths.
    :param rm_err: boolean specifying whether to delete the file if an error is found
    :param print_err: print any errors associated with the file to the console
    :param print_file_names: print the names of the file before they are tested.
    :param timeout: a timeout in seconds (Default = 4) for the tests to be undertaken.
    :return: boolean specifying whether the file is OK (i.e., tests passed) or not.

    """
    from multiprocessing import Pool

    files_ok = True
    try:
        processes_pool = Pool(1)
        for input_file in input_files:
            if print_file_names:
                print(input_file)
            params = [input_file, rm_err, print_err]
            try:
                result = processes_pool.apply_async(_run_check_hdf5_file, args=[params])
                file_ok = result.get(timeout=timeout)
                if not file_ok:
                    files_ok = False
            except Exception as e:
                files_ok = False
                if rm_err:
                    rsgislib.tools.filetools.delete_file_silent(input_file)
                    print("Removed {}".format(input_file))
                else:
                    print("rm {}".format(input_file))
        processes_pool.close()
        processes_pool.join()
    except Exception as inst:
        print("Finished with pool")

    return files_ok



import os
import pytest

PYOD_NOT_AVAIL = False
try:
    import pyod
except ImportError:
    PYOD_NOT_AVAIL = True

MATPLOTLIB_NOT_AVAIL = False
try:
    import matplotlib.pyplot
except ImportError:
    MATPLOTLIB_NOT_AVAIL = True

DATA_DIR = os.path.join(os.path.dirname(os.path.abspath(__file__)), "data")
CHANGEDETECT_DATA_DIR = os.path.join(DATA_DIR, "changedetect")


@pytest.mark.skipif(PYOD_NOT_AVAIL, reason="pyod dependency not available")
def test_find_class_pyod_outliers(tmp_path):
    import rsgislib.changedetect.pxloutlierchng
    import pyod.models.knn

    pyod_obj = pyod.models.knn.KNN()
    input_img = os.path.join(CHANGEDETECT_DATA_DIR, "LS8_20180608_ndvi_sub.kea")
    in_msk_img = os.path.join(CHANGEDETECT_DATA_DIR, "base_1997_class_img_sub.kea")
    output_img = os.path.join(tmp_path, "out_img.kea")
    out_scores_img = os.path.join(tmp_path, "out_score_img.kea")

    rsgislib.changedetect.pxloutlierchng.find_class_pyod_outliers(
        pyod_obj,
        input_img,
        in_msk_img,
        output_img,
        out_scores_img=out_scores_img,
        img_mask_val=1,
        img_bands=None,
        gdalformat="KEA",
    )

    assert os.path.exists(output_img) and os.path.exists(out_scores_img)


@pytest.mark.skipif(MATPLOTLIB_NOT_AVAIL, reason="matplotlib dependency not available")
def test_find_class_kurt_skew_outliers(tmp_path):
    from rsgislib.changedetect.pxloutlierchng import find_class_kurt_skew_outliers

    input_img = os.path.join(CHANGEDETECT_DATA_DIR, "LS8_20180608_ndvi_sub.kea")
    in_msk_img = os.path.join(CHANGEDETECT_DATA_DIR, "base_1997_class_img_sub.kea")
    output_img = os.path.join(tmp_path, "out_img.kea")
    plot_thres_file = os.path.join(tmp_path, "out_plot.png")

    find_class_kurt_skew_outliers(
        input_img,
        in_msk_img,
        output_img,
        vld_min=-1,
        vld_max=1,
        init_thres=0.25,
        low_thres=True,
        contamination=10.0,
        only_kurtosis=False,
        img_mask_val=1,
        img_band=1,
        img_val_no_data=-999,
        gdalformat="KEA",
        plot_thres_file=plot_thres_file,
    )

    assert os.path.exists(output_img)


@pytest.mark.skipif(MATPLOTLIB_NOT_AVAIL, reason="matplotlib dependency not available")
def test_find_class_otsu_outliers(tmp_path):
    from rsgislib.changedetect.pxloutlierchng import find_class_otsu_outliers

    input_img = os.path.join(CHANGEDETECT_DATA_DIR, "LS8_20180608_ndvi_sub.kea")
    in_msk_img = os.path.join(CHANGEDETECT_DATA_DIR, "base_1997_class_img_sub.kea")
    output_img = os.path.join(tmp_path, "out_img.kea")
    plot_thres_file = os.path.join(tmp_path, "out_plot.png")

    find_class_otsu_outliers(
        input_img,
        in_msk_img,
        output_img,
        low_thres=True,
        img_mask_val=1,
        img_band=1,
        img_val_no_data=-999,
        gdalformat="KEA",
        plot_thres_file=plot_thres_file,
    )

    assert os.path.exists(output_img)


@pytest.mark.skipif(MATPLOTLIB_NOT_AVAIL, reason="matplotlib dependency not available")
def test_find_class_li_outliers(tmp_path):
    from rsgislib.changedetect.pxloutlierchng import find_class_li_outliers

    input_img = os.path.join(CHANGEDETECT_DATA_DIR, "LS8_20180608_ndvi_sub.kea")
    in_msk_img = os.path.join(CHANGEDETECT_DATA_DIR, "base_1997_class_img_sub.kea")
    output_img = os.path.join(tmp_path, "out_img.kea")
    plot_thres_file = os.path.join(tmp_path, "out_plot.png")

    find_class_li_outliers(
        input_img,
        in_msk_img,
        output_img,
        low_thres=True,
        tolerance=None,
        init_thres=0.35,
        img_mask_val=1,
        img_band=1,
        img_val_no_data=-999,
        gdalformat="KEA",
        plot_thres_file=plot_thres_file,
    )

    assert os.path.exists(output_img)



#! /usr/bin/env python
############################################################################
#  tiledfilter.py
#
#  Copyright 2016 RSGISLib.
#
#  RSGISLib: 'The Remote Sensing and GIS Software Library'
#
#  RSGISLib is free software: you can redistribute it and/or modify
#  it under the terms of the GNU General Public License as published by
#  the Free Software Foundation, either version 3 of the License, or
#  (at your option) any later version.
#
#  RSGISLib is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#
#  You should have received a copy of the GNU General Public License
#  along with RSGISLib.  If not, see <http://www.gnu.org/licenses/>.
#
#
# Purpose:  Provide functionality to perform image filtering using tiling
#           and multiple processing cores.
#
# Author: Pete Bunting
# Email: petebunting@mac.com
# Date: 14/05/2016
# Version: 1.0
#
# History:
# Version 1.0 - Created.
#
###########################################################################

import rsgislib
from rsgislib import imageutils
from rsgislib import imagefilter

import glob
import os.path
import os
import shutil
import math
from multiprocessing import Pool
from abc import ABCMeta, abstractmethod


class RSGISAbstractFilter:
    """
    Abstract class for filter defining the interface to be used within the
    perform_tiled_img_filter function.

    :param self.filter_size: size of the image filter (must be an odd number)
    :param self.gdalformat: the output image file format
    :param self.datatype: the output image data type (e.g., rsgislib.TYPE_16UINT)
    """

    __metaclass__ = ABCMeta

    def __init__(self):
        self.filter_size = 3
        self.gdalformat = "KEA"
        self.datatype = rsgislib.TYPE_32FLOAT

    @abstractmethod
    def applyFilter(self, input_img, output_img):
        """
        Abstract function through which the input image
        is filtered to produce the output image.

        :param input_img: is the name and path of the input image.
        :param output_img: is the name and path of the output image.
        """
        pass

    def getFilterSize(self):
        """
        return the size of the image filter
        """
        return self.filter_size

    def getFilterHSize(self):
        """
        return the hald size of the image filter (i.e., (filter_size-1)/2)
        """
        hSize = (self.filter_size - 1) / 2
        return hSize


class RSGISMedianFilter(RSGISAbstractFilter):
    """A class to apply a median filter"""

    def __init__(self, filter_size, gdalformat, datatype):
        self.filter_size = filter_size
        self.gdalformat = gdalformat
        self.datatype = datatype

    def applyFilter(self, input_img, output_img):
        """Apply a median filter to the specified input image."""
        out_image_base, out_img_ext = os.path.splitext(output_img)
        out_img_ext = out_img_ext.replace(".", "").strip()
        filters = []
        filters.append(
            imagefilter.FilterParameters(
                filter_type="Median", file_ending="", size=self.filter_size
            )
        )
        imagefilter.apply_filters(
            input_img,
            out_image_base,
            filters,
            self.gdalformat,
            out_img_ext,
            self.datatype,
        )


class RSGISMeanFilter(RSGISAbstractFilter):
    """A class to apply a mean filter"""

    def __init__(self, filter_size, gdalformat, datatype):
        self.filter_size = filter_size
        self.gdalformat = gdalformat
        self.datatype = datatype

    def applyFilter(self, input_img, output_img):
        """Apply a mean filter to the specified input image."""
        out_image_base, out_img_ext = os.path.splitext(output_img)
        out_img_ext = out_img_ext.replace(".", "").strip()
        filters = []
        filters.append(
            imagefilter.FilterParameters(
                filter_type="Mean", file_ending="", size=self.filter_size
            )
        )
        imagefilter.apply_filters(
            input_img,
            out_image_base,
            filters,
            self.gdalformat,
            out_img_ext,
            self.datatype,
        )


class RSGISMinFilter(RSGISAbstractFilter):
    """A class to apply a min filter"""

    def __init__(self, filter_size, gdalformat, datatype):
        self.filter_size = filter_size
        self.gdalformat = gdalformat
        self.datatype = datatype

    def applyFilter(self, input_img, output_img):
        """Apply a min filter to the specified input image."""
        out_image_base, out_img_ext = os.path.splitext(output_img)
        out_img_ext = out_img_ext.replace(".", "").strip()
        filters = []
        filters.append(
            imagefilter.FilterParameters(
                filter_type="Min", file_ending="", size=self.filter_size
            )
        )
        imagefilter.apply_filters(
            input_img,
            out_image_base,
            filters,
            self.gdalformat,
            out_img_ext,
            self.datatype,
        )


class RSGISMaxFilter(RSGISAbstractFilter):
    """A class to apply a max filter"""

    def __init__(self, filter_size, gdalformat, datatype):
        self.filter_size = filter_size
        self.gdalformat = gdalformat
        self.datatype = datatype

    def applyFilter(self, input_img, output_img):
        """Apply a max filter to the specified input image."""
        out_image_base, out_img_ext = os.path.splitext(output_img)
        out_img_ext = out_img_ext.replace(".", "").strip()
        filters = []
        filters.append(
            imagefilter.FilterParameters(
                filter_type="Max", file_ending="", size=self.filter_size
            )
        )
        imagefilter.apply_filters(
            input_img,
            out_image_base,
            filters,
            self.gdalformat,
            out_img_ext,
            self.datatype,
        )


class RSGISModeFilter(RSGISAbstractFilter):
    """A class to apply a mode filter"""

    def __init__(self, filter_size, gdalformat, datatype):
        self.filter_size = filter_size
        self.gdalformat = gdalformat
        self.datatype = datatype

    def applyFilter(self, input_img, output_img):
        """Apply a mode filter to the specified input image."""
        out_image_base, out_img_ext = os.path.splitext(output_img)
        out_img_ext = out_img_ext.replace(".", "").strip()
        filters = []
        filters.append(
            imagefilter.FilterParameters(
                filter_type="Mode", file_ending="", size=self.filter_size
            )
        )
        imagefilter.apply_filters(
            input_img,
            out_image_base,
            filters,
            self.gdalformat,
            out_img_ext,
            self.datatype,
        )


class RSGISStdDevFilter(RSGISAbstractFilter):
    """A class to apply a standard deviation filter"""

    def __init__(self, filter_size, gdalformat, datatype):
        self.filter_size = filter_size
        self.gdalformat = gdalformat
        self.datatype = datatype

    def applyFilter(self, input_img, output_img):
        """Apply a std dev filter to the specified input image."""
        out_image_base, out_img_ext = os.path.splitext(output_img)
        out_img_ext = out_img_ext.replace(".", "").strip()
        filters = []
        filters.append(
            imagefilter.FilterParameters(
                filter_type="StdDev", file_ending="", size=self.filter_size
            )
        )
        imagefilter.apply_filters(
            input_img,
            out_image_base,
            filters,
            self.gdalformat,
            out_img_ext,
            self.datatype,
        )


class RSGISRangeFilter(RSGISAbstractFilter):
    """A class to apply a range filter"""

    def __init__(self, filter_size, gdalformat, datatype):
        self.filter_size = filter_size
        self.gdalformat = gdalformat
        self.datatype = datatype

    def applyFilter(self, input_img, output_img):
        """Apply a range filter to the specified input image."""
        out_image_base, out_img_ext = os.path.splitext(output_img)
        out_img_ext = out_img_ext.replace(".", "").strip()
        filters = []
        filters.append(
            imagefilter.FilterParameters(
                filter_type="Range", file_ending="", size=self.filter_size
            )
        )
        imagefilter.apply_filters(
            input_img,
            out_image_base,
            filters,
            self.gdalformat,
            out_img_ext,
            self.datatype,
        )


class RSGISMeanDiffFilter(RSGISAbstractFilter):
    """A class to apply a mean difference filter"""

    def __init__(self, filter_size, gdalformat, datatype):
        self.filter_size = filter_size
        self.gdalformat = gdalformat
        self.datatype = datatype

    def applyFilter(self, input_img, output_img):
        """Apply a mean difference filter to the specified input image."""
        out_image_base, out_img_ext = os.path.splitext(output_img)
        out_img_ext = out_img_ext.replace(".", "").strip()
        filters = []
        filters.append(
            imagefilter.FilterParameters(
                filter_type="MeanDiff", file_ending="", size=self.filter_size
            )
        )
        imagefilter.apply_filters(
            input_img,
            out_image_base,
            filters,
            self.gdalformat,
            out_img_ext,
            self.datatype,
        )


class RSGISMeanDiffAbsFilter(RSGISAbstractFilter):
    """A class to apply a mean absolute difference filter"""

    def __init__(self, filter_size, gdalformat, datatype):
        self.filter_size = filter_size
        self.gdalformat = gdalformat
        self.datatype = datatype

    def applyFilter(self, input_img, output_img):
        """Apply a mean absolute difference filter to the specified input image."""
        out_image_base, out_img_ext = os.path.splitext(output_img)
        out_img_ext = out_img_ext.replace(".", "").strip()
        filters = []
        filters.append(
            imagefilter.FilterParameters(
                filter_type="MeanDiffAbs", file_ending="", size=self.filter_size
            )
        )
        imagefilter.apply_filters(
            input_img,
            out_image_base,
            filters,
            self.gdalformat,
            out_img_ext,
            self.datatype,
        )


class RSGISTotalDiffFilter(RSGISAbstractFilter):
    """A class to apply a total difference filter"""

    def __init__(self, filter_size, gdalformat, datatype):
        self.filter_size = filter_size
        self.gdalformat = gdalformat
        self.datatype = datatype

    def applyFilter(self, input_img, output_img):
        """Apply a total (i.e., sum) difference filter to the specified input image."""
        out_image_base, out_img_ext = os.path.splitext(output_img)
        out_img_ext = out_img_ext.replace(".", "").strip()
        filters = []
        filters.append(
            imagefilter.FilterParameters(
                filter_type="TotalDiff", file_ending="", size=self.filter_size
            )
        )
        imagefilter.apply_filters(
            input_img,
            out_image_base,
            filters,
            self.gdalformat,
            out_img_ext,
            self.datatype,
        )


class RSGISTotalDiffAbsFilter(RSGISAbstractFilter):
    """A class to apply a total absolution difference filter"""

    def __init__(self, filter_size, gdalformat, datatype):
        self.filter_size = filter_size
        self.gdalformat = gdalformat
        self.datatype = datatype

    def applyFilter(self, input_img, output_img):
        """Apply a total absolute difference filter to the specified input image."""
        out_image_base, out_img_ext = os.path.splitext(output_img)
        out_img_ext = out_img_ext.replace(".", "").strip()
        filters = []
        filters.append(
            imagefilter.FilterParameters(
                filter_type="TotalDiffAbs", file_ending="", size=self.filter_size
            )
        )
        imagefilter.apply_filters(
            input_img,
            out_image_base,
            filters,
            self.gdalformat,
            out_img_ext,
            self.datatype,
        )


class RSGISCoeffOfVarFilter(RSGISAbstractFilter):
    """A class to apply a coefficient of variance filter"""

    def __init__(self, filter_size, gdalformat, datatype):
        self.filter_size = filter_size
        self.gdalformat = gdalformat
        self.datatype = datatype

    def applyFilter(self, input_img, output_img):
        """Apply a coefficient of variance filter to the specified input image."""
        out_image_base, out_img_ext = os.path.splitext(output_img)
        out_img_ext = out_img_ext.replace(".", "").strip()
        filters = []
        filters.append(
            imagefilter.FilterParameters(
                filter_type="CoeffOfVar", file_ending="", size=self.filter_size
            )
        )
        imagefilter.apply_filters(
            input_img,
            out_image_base,
            filters,
            self.gdalformat,
            out_img_ext,
            self.datatype,
        )


class RSGISTotalFilter(RSGISAbstractFilter):
    """A class to apply a total filter"""

    def __init__(self, filter_size, gdalformat, datatype):
        self.filter_size = filter_size
        self.gdalformat = gdalformat
        self.datatype = datatype

    def applyFilter(self, input_img, output_img):
        """Apply a total filter to the specified input image."""
        out_image_base, out_img_ext = os.path.splitext(output_img)
        out_img_ext = out_img_ext.replace(".", "").strip()
        filters = []
        filters.append(
            imagefilter.FilterParameters(
                filter_type="Total", file_ending="", size=self.filter_size
            )
        )
        imagefilter.apply_filters(
            input_img,
            out_image_base,
            filters,
            self.gdalformat,
            out_img_ext,
            self.datatype,
        )


class RSGISNormVarFilter(RSGISAbstractFilter):
    """A class to apply a normalised variance filter"""

    def __init__(self, filter_size, gdalformat, datatype):
        self.filter_size = filter_size
        self.gdalformat = gdalformat
        self.datatype = datatype

    def applyFilter(self, input_img, output_img):
        """Apply a normalised variance filter to the specified input image."""
        out_image_base, out_img_ext = os.path.splitext(output_img)
        out_img_ext = out_img_ext.replace(".", "").strip()
        filters = []
        filters.append(
            imagefilter.FilterParameters(
                filter_type="NormVar", file_ending="", size=self.filter_size
            )
        )
        imagefilter.apply_filters(
            input_img,
            out_image_base,
            filters,
            self.gdalformat,
            out_img_ext,
            self.datatype,
        )


class RSGISNormVarSqrtFilter(RSGISAbstractFilter):
    """A class to apply a normalised variance square root filter"""

    def __init__(self, filter_size, gdalformat, datatype):
        self.filter_size = filter_size
        self.gdalformat = gdalformat
        self.datatype = datatype

    def applyFilter(self, input_img, output_img):
        """
        Apply a normalised variance square root filter to the specified input image.
        """
        out_image_base, out_img_ext = os.path.splitext(output_img)
        out_img_ext = out_img_ext.replace(".", "").strip()
        filters = []
        filters.append(
            imagefilter.FilterParameters(
                filter_type="NormVarSqrt", file_ending="", size=self.filter_size
            )
        )
        imagefilter.apply_filters(
            input_img,
            out_image_base,
            filters,
            self.gdalformat,
            out_img_ext,
            self.datatype,
        )


class RSGISNormVarLnFilter(RSGISAbstractFilter):
    """A class to apply a normalised variance log filter"""

    def __init__(self, filter_size, gdalformat, datatype):
        self.filter_size = filter_size
        self.gdalformat = gdalformat
        self.datatype = datatype

    def applyFilter(self, input_img, output_img):
        """
        Apply a normalised variance natural log filter to the specified input image.
        """
        out_image_base, out_img_ext = os.path.splitext(output_img)
        out_img_ext = out_img_ext.replace(".", "").strip()
        filters = []
        filters.append(
            imagefilter.FilterParameters(
                filter_type="NormVarLn", file_ending="", size=self.filter_size
            )
        )
        imagefilter.apply_filters(
            input_img,
            out_image_base,
            filters,
            self.gdalformat,
            out_img_ext,
            self.datatype,
        )


class RSGISTextureVarFilter(RSGISAbstractFilter):
    """A class to apply a texture variance filter"""

    def __init__(self, filter_size, gdalformat, datatype):
        self.filter_size = filter_size
        self.gdalformat = gdalformat
        self.datatype = datatype

    def applyFilter(self, input_img, output_img):
        """Apply a texture variance filter to the specified input image."""
        out_image_base, out_img_ext = os.path.splitext(output_img)
        out_img_ext = out_img_ext.replace(".", "").strip()
        filters = []
        filters.append(
            imagefilter.FilterParameters(
                filter_type="TextureVar", file_ending="", size=self.filter_size
            )
        )
        imagefilter.apply_filters(
            input_img,
            out_image_base,
            filters,
            self.gdalformat,
            out_img_ext,
            self.datatype,
        )


class RSGISKuwaharaFilter(RSGISAbstractFilter):
    """A class to apply a kuwahara filter"""

    def __init__(self, filter_size, gdalformat, datatype):
        self.filter_size = filter_size
        self.gdalformat = gdalformat
        self.datatype = datatype

    def applyFilter(self, input_img, output_img):
        """Apply a kuwahara filter to the specified input image."""
        out_image_base, out_img_ext = os.path.splitext(output_img)
        out_img_ext = out_img_ext.replace(".", "").strip()
        filters = []
        filters.append(
            imagefilter.FilterParameters(
                filter_type="Kuwahara", file_ending="", size=self.filter_size
            )
        )
        imagefilter.apply_filters(
            input_img,
            out_image_base,
            filters,
            self.gdalformat,
            out_img_ext,
            self.datatype,
        )


class RSGISSobelFilter(RSGISAbstractFilter):
    """A class to apply a sobel filter"""

    def __init__(self, gdalformat, datatype):
        self.filter_size = 3
        self.gdalformat = gdalformat
        self.datatype = datatype

    def applyFilter(self, input_img, output_img):
        """Apply a sobel filter to the specified input image."""
        out_image_base, out_img_ext = os.path.splitext(output_img)
        out_img_ext = out_img_ext.replace(".", "").strip()
        filters = []
        filters.append(
            imagefilter.FilterParameters(
                filter_type="Sobel", file_ending="", option="xy"
            )
        )
        imagefilter.apply_filters(
            input_img,
            out_image_base,
            filters,
            self.gdalformat,
            out_img_ext,
            self.datatype,
        )


class RSGISSobelXFilter(RSGISAbstractFilter):
    """A class to apply a sobel X filter"""

    def __init__(self, gdalformat, datatype):
        self.filter_size = 3
        self.gdalformat = gdalformat
        self.datatype = datatype

    def applyFilter(self, input_img, output_img):
        """Apply a sobel filter in X axis to the specified input image."""
        out_image_base, out_img_ext = os.path.splitext(output_img)
        out_img_ext = out_img_ext.replace(".", "").strip()
        filters = []
        filters.append(
            imagefilter.FilterParameters(
                filter_type="Sobel", file_ending="", option="x"
            )
        )
        imagefilter.apply_filters(
            input_img,
            out_image_base,
            filters,
            self.gdalformat,
            out_img_ext,
            self.datatype,
        )


class RSGISSobelYFilter(RSGISAbstractFilter):
    """A class to apply a sobel Y filter"""

    def __init__(self, gdalformat, datatype):
        self.filter_size = 3
        self.gdalformat = gdalformat
        self.datatype = datatype

    def applyFilter(self, input_img, output_img):
        """Apply a sobel filter in Y axis to the specified input image."""
        out_image_base, out_img_ext = os.path.splitext(output_img)
        out_img_ext = out_img_ext.replace(".", "").strip()
        filters = []
        filters.append(
            imagefilter.FilterParameters(
                filter_type="Sobel", file_ending="", option="y"
            )
        )
        imagefilter.apply_filters(
            input_img,
            out_image_base,
            filters,
            self.gdalformat,
            out_img_ext,
            self.datatype,
        )


class RSGISPrewittFilter(RSGISAbstractFilter):
    """A class to apply a Prewitt filter"""

    def __init__(self, gdalformat, datatype):
        self.filter_size = 3
        self.gdalformat = gdalformat
        self.datatype = datatype

    def applyFilter(self, input_img, output_img):
        """Apply a Prewitt filter to the specified input image."""
        out_image_base, out_img_ext = os.path.splitext(output_img)
        out_img_ext = out_img_ext.replace(".", "").strip()
        filters = []
        filters.append(
            imagefilter.FilterParameters(
                filter_type="Prewitt", file_ending="", option="xy"
            )
        )
        imagefilter.apply_filters(
            input_img,
            out_image_base,
            filters,
            self.gdalformat,
            out_img_ext,
            self.datatype,
        )


class RSGISPrewittXFilter(RSGISAbstractFilter):
    """A class to apply a Prewitt X filter"""

    def __init__(self, gdalformat, datatype):
        self.filter_size = 3
        self.gdalformat = gdalformat
        self.datatype = datatype

    def applyFilter(self, input_img, output_img):
        """Apply a Prewitt filter in X axis to the specified input image."""
        out_image_base, out_img_ext = os.path.splitext(output_img)
        out_img_ext = out_img_ext.replace(".", "").strip()
        filters = []
        filters.append(
            imagefilter.FilterParameters(
                filter_type="Prewitt", file_ending="", option="x"
            )
        )
        imagefilter.apply_filters(
            input_img,
            out_image_base,
            filters,
            self.gdalformat,
            out_img_ext,
            self.datatype,
        )


class RSGISPrewittYFilter(RSGISAbstractFilter):
    """A class to apply a Prewitt Y filter"""

    def __init__(self, gdalformat, datatype):
        self.filter_size = 3
        self.gdalformat = gdalformat
        self.datatype = datatype

    def applyFilter(self, input_img, output_img):
        """Apply a Prewitt filter in Y axis to the specified input image."""
        out_image_base, out_img_ext = os.path.splitext(output_img)
        out_img_ext = out_img_ext.replace(".", "").strip()
        filters = []
        filters.append(
            imagefilter.FilterParameters(
                filter_type="Prewitt", file_ending="", option="y"
            )
        )
        imagefilter.apply_filters(
            input_img,
            out_image_base,
            filters,
            self.gdalformat,
            out_img_ext,
            self.datatype,
        )


class RSGISGaussianSmoothFilter(RSGISAbstractFilter):
    """A class to apply a Gaussian smoothing filter"""

    def __init__(
        self, filter_size, stddev_x, stddev_y, filter_angle, gdalformat, datatype
    ):
        self.filter_size = filter_size
        self.stddev_x = stddev_x
        self.stddev_y = stddev_y
        self.filter_angle = filter_angle
        self.gdalformat = gdalformat
        self.datatype = datatype

    def applyFilter(self, input_img, output_img):
        """Apply a Gaussian smoothing filter to the specified input image."""
        out_image_base, out_img_ext = os.path.splitext(output_img)
        out_img_ext = out_img_ext.replace(".", "").strip()
        filters = []
        filters.append(
            imagefilter.FilterParameters(
                filter_type="GaussianSmooth",
                file_ending="",
                size=self.filter_size,
                stddev_x=self.stddev_x,
                stddev_y=self.stddev_y,
                angle=self.filter_angle,
            )
        )
        imagefilter.apply_filters(
            input_img,
            out_image_base,
            filters,
            self.gdalformat,
            out_img_ext,
            self.datatype,
        )


class RSGISGaussian1stDerivFilter(RSGISAbstractFilter):
    """A class to apply a Gaussian first derivative filter"""

    def __init__(
        self, filter_size, stddev_x, stddev_y, filter_angle, gdalformat, datatype
    ):
        self.filter_size = filter_size
        self.stddev_x = stddev_x
        self.stddev_y = stddev_y
        self.filter_angle = filter_angle
        self.gdalformat = gdalformat
        self.datatype = datatype

    def applyFilter(self, input_img, output_img):
        """Apply a Gaussian first derivative filter to the specified input image."""
        out_image_base, out_img_ext = os.path.splitext(output_img)
        out_img_ext = out_img_ext.replace(".", "").strip()
        filters = []
        filters.append(
            imagefilter.FilterParameters(
                filter_type="Gaussian1st",
                file_ending="",
                size=self.filter_size,
                stddev_x=self.stddev_x,
                stddev_y=self.stddev_y,
                angle=self.filter_angle,
            )
        )
        imagefilter.apply_filters(
            input_img,
            out_image_base,
            filters,
            self.gdalformat,
            out_img_ext,
            self.datatype,
        )


class RSGISGaussian2ndDerivFilter(RSGISAbstractFilter):
    """A class to apply a Gaussian second derivative filter"""

    def __init__(
        self, filter_size, stddev_x, stddev_y, filter_angle, gdalformat, datatype
    ):
        self.filter_size = filter_size
        self.stddev_x = stddev_x
        self.stddev_y = stddev_y
        self.filter_angle = filter_angle
        self.gdalformat = gdalformat
        self.datatype = datatype

    def applyFilter(self, input_img, output_img):
        """Apply a Gaussian second derivative filter to the specified input image."""
        out_image_base, out_img_ext = os.path.splitext(output_img)
        out_img_ext = out_img_ext.replace(".", "").strip()
        filters = []
        filters.append(
            imagefilter.FilterParameters(
                filter_type="Gaussian2nd",
                file_ending="",
                size=self.filter_size,
                stddev_x=self.stddev_x,
                stddev_y=self.stddev_y,
                angle=self.filter_angle,
            )
        )
        imagefilter.apply_filters(
            input_img,
            out_image_base,
            filters,
            self.gdalformat,
            out_img_ext,
            self.datatype,
        )


class RSGISLaplacianFilter(RSGISAbstractFilter):
    """A class to apply a Laplacian filter"""

    def __init__(self, filter_size, stddev, gdalformat, datatype):
        self.filter_size = filter_size
        self.stddev = stddev
        self.gdalformat = gdalformat
        self.datatype = datatype

    def applyFilter(self, input_img, output_img):
        """Apply a Laplacian filter to the specified input image."""
        out_image_base, out_img_ext = os.path.splitext(output_img)
        out_img_ext = out_img_ext.replace(".", "").strip()
        filters = []
        filters.append(
            imagefilter.FilterParameters(
                filter_type="Laplacian",
                file_ending="",
                size=self.filter_size,
                stddev=self.stddev,
            )
        )
        imagefilter.apply_filters(
            input_img,
            out_image_base,
            filters,
            self.gdalformat,
            out_img_ext,
            self.datatype,
        )


class RSGISLeeFilter(RSGISAbstractFilter):
    """A class to apply a SAR Lee filter"""

    def __init__(self, filter_size, n_looks, gdalformat, datatype):
        self.filter_size = filter_size
        self.n_looks = n_looks
        self.gdalformat = gdalformat
        self.datatype = datatype

    def applyFilter(self, input_img, output_img):
        """Apply a Lee SAR filter to the specified input image."""
        out_image_base, out_img_ext = os.path.splitext(output_img)
        out_img_ext = out_img_ext.replace(".", "").strip()
        filters = []
        filters.append(
            imagefilter.FilterParameters(
                filter_type="Lee",
                file_ending="",
                size=self.filter_size,
                n_looks=self.n_looks,
            )
        )
        imagefilter.apply_filters(
            input_img,
            out_image_base,
            filters,
            self.gdalformat,
            out_img_ext,
            self.datatype,
        )


def _perform_filtering_func(filter_params):
    """
    Clump an image with values provides as an array for use within a
    multiprocessing Pool
    """
    filter_params[2].applyFilter(filter_params[0], filter_params[1])


def perform_tiled_img_filter(
    input_img,
    output_img,
    filter_inst,
    datatype=None,
    gdalformat="KEA",
    tmp_dir="tmp",
    width=2000,
    height=2000,
    n_cores=-1,
):
    """
    This function will perform a filtering of an input image where the input image
    will be tiled and the tiles executed on multiple processing cores. This function
    is primarily of use for larger images or when using very large filter windows
    otherwise the over head of tiling and mosaicking are not worth it.

    :param input_img: is the file name and path for the input image file.
    :param output_img: is the file name and path for the output image file.
    :param filter_inst: is an instance of a filter class available within
                        rsgislib.imagefilter.tiledfilter.
    :param datatype: is the output image data type (e.g., rsgislib.TYPE_32FLOAT;
                     Default is None). If None then data type of input image is used.
    :param gdalformat: string with the GDAL image format for the output image
                       (Default = KEA). NOTE. KEA is used as intermediate format
                       internally and therefore needs to be available.
    :param tmp_dir: the temporary directory where intermediate files will be written
                    (default is 'tmp'). Directory will be created and deleted if does
                    not exist.
    :param width: int for width of the image tiles used for processing (Default = 2000).
    :param height: int for height of the image tiles used for processing
                   (Default = 2000).
    :param n_cores: is an int specifying the number of cores to be used for processing.

    .. code:: python

        import rsgislib
        from rsgislib.imagefilter import tiledfilter
        from rsgislib import imageutils

        input_img = 'LandsatImg.kea'
        output_img = 'LandsatImgMedianFilter.kea'

        medianFilter = tiledfilter.RSGISMedianFilter(7, "KEA", rsgislib.TYPE_16UINT)
        tiledfilter.perform_tiled_img_filter(input_img, output_img, medianFilter,
                                             width=2000, height=2000)
        imageutils.pop_img_stats(output_img, False, 0, True)

    """
    import rsgislib.tools.utils

    createdTmp = False
    if not os.path.exists(tmp_dir):
        os.makedirs(tmp_dir)
        createdTmp = True

    if n_cores <= 0:
        n_cores = rsgislib.tools.utils.num_process_cores()

    uidStr = rsgislib.tools.utils.uid_generator()
    if datatype is None:
        datatype = rsgislib.imageutils.get_rsgislib_datatype_from_img(input_img)

    baseName = os.path.splitext(os.path.basename(input_img))[0] + "_" + uidStr

    imgTilesDIR = os.path.join(tmp_dir, "imgTiles_" + uidStr)
    tilesFilterDIR = os.path.join(tmp_dir, "imgFilterTiles_" + uidStr)
    tilesImgBase = os.path.join(imgTilesDIR, baseName)

    if not os.path.exists(imgTilesDIR):
        os.makedirs(imgTilesDIR)

    if not os.path.exists(tilesFilterDIR):
        os.makedirs(tilesFilterDIR)

    tileOverlap = math.ceil(filter_inst.getFilterHSize())

    imageutils.create_tiles(
        input_img,
        tilesImgBase,
        int(width),
        int(height),
        int(tileOverlap),
        False,
        "KEA",
        datatype,
        "kea",
    )
    imageTiles = glob.glob(tilesImgBase + "*.kea")

    filterImgsVals = []
    for tile in imageTiles:
        tileBaseName = os.path.splitext(os.path.basename(tile))[0]
        filterTile = os.path.join(tilesFilterDIR, tileBaseName + "_filter.kea")
        filterImgsVals.append([tile, filterTile, filter_inst])

    with Pool(n_cores) as p:
        p.map(_perform_filtering_func, filterImgsVals)

    imgFilterTiles = glob.glob(os.path.join(tilesFilterDIR, "*_filter.kea"))

    numOutBands = rsgislib.imageutils.get_img_band_count(input_img)

    imageutils.create_copy_img(
        input_img, output_img, numOutBands, 0, gdalformat, datatype
    )

    imageutils.include_imgs_with_overlap(output_img, imgFilterTiles, int(tileOverlap))

    shutil.rmtree(imgTilesDIR)
    shutil.rmtree(tilesFilterDIR)
    if createdTmp:
        shutil.rmtree(tmp_dir)


def perform_tiled_img_multi_filter(
    input_img,
    output_imgs,
    filter_insts,
    datatype=None,
    gdalformat="KEA",
    tmp_dir="tmp",
    width=2000,
    height=2000,
    n_cores=-1,
):
    """
    This function will perform the filtering using multiple filters of an input image
    where the input image will be tiled and the tiles executed on multiple processing
    cores. This function is primarily of use for larger images or when using very large
    filter windows otherwise the over head of tiling and mosaicking are not worth it.

    :param input_img: is the file name and path for the input image file.
    :param output_imgs: is a list of file names and paths for the output image files -
                       Note, must be the same length as filter_insts.
    :param filter_insts: is a list of filter instances of the classes available
                         within rsgislib.imagefilter.tiledfilter  - Note, must be the
                         same length as filter_insts.
    :param datatype: is the output image data type (e.g., rsgislib.TYPE_32FLOAT;
                     Default is None). If None then data type of input image is used.
    :param gdalformat: string with the GDAL image format for the output image
                       (Default = KEA). NOTE. KEA is used as intermediate format
                       internally and therefore needs to be available.
    :param tmp_dir: the temporary directory where intermediate files will be
                    written (default is 'tmp'). Directory will be created and
                    deleted if does not exist.
    :param width: int for width of the image tiles used for processing (Default = 2000).
    :param height: int for height of the image tiles used for processing
                   (Default = 2000).
    :param n_cores: is an int specifying the number of cores to be used for processing.

    .. code:: python

        import rsgislib
        from rsgislib.imagefilter import tiledfilter
        from rsgislib import imageutils

        input_img = 'LandsatImg.kea'
        outputImages = ['LandsatImgMedianFilter.kea', 'LandsatImgNormVarFilter.kea']

        filters = [tiledfilter.RSGISMedianFilter(7, "KEA", rsgislib.TYPE_16UINT),
                  tiledfilter.RSGISNormVarFilter(7, "KEA", rsgislib.TYPE_16UINT]
        tiledfilter.perform_tiled_img_multi_filter(input_img, outputImages,
                                                   filters, width=2000, height=2000)
        imageutils.pop_img_stats(output_img, False, 0, True)

    """
    import rsgislib.tools.utils

    if len(output_imgs) != len(filter_insts):
        raise rsgislib.RSGISPyException(
            "The same number of filters and output images need to be provided."
        )

    numFilters = len(output_imgs)

    createdTmp = False
    if not os.path.exists(tmp_dir):
        os.makedirs(tmp_dir)
        createdTmp = True

    if n_cores <= 0:
        n_cores = rsgislib.tools.utils.num_process_cores()

    uidStr = rsgislib.tools.utils.uid_generator()
    if datatype is None:
        datatype = rsgislib.imageutils.get_rsgislib_datatype_from_img(input_img)

    baseName = os.path.splitext(os.path.basename(input_img))[0] + "_" + uidStr

    imgTilesDIR = os.path.join(tmp_dir, "imgTiles_" + uidStr)
    tilesFilterDIR = os.path.join(tmp_dir, "imgFilterTiles_" + uidStr)
    tilesImgBase = os.path.join(imgTilesDIR, baseName)

    if not os.path.exists(imgTilesDIR):
        os.makedirs(imgTilesDIR)

    first = True
    tileOverlap = 0

    for filter_inst in filter_insts:
        tmpOverlap = math.ceil(filter_inst.getFilterHSize())
        if first:
            tileOverlap = tmpOverlap
            first = False
        elif tmpOverlap > tileOverlap:
            tileOverlap = tmpOverlap

    imageutils.create_tiles(
        input_img,
        tilesImgBase,
        int(width),
        int(height),
        int(tileOverlap),
        False,
        "KEA",
        datatype,
        "kea",
    )
    imageTiles = glob.glob(tilesImgBase + "*.kea")

    for i in range(numFilters):
        filter_inst = filter_insts[i]
        output_img = output_imgs[i]

        if not os.path.exists(tilesFilterDIR):
            os.makedirs(tilesFilterDIR)

        filterImgsVals = []
        for tile in imageTiles:
            tileBaseName = os.path.splitext(os.path.basename(tile))[0]
            filterTile = os.path.join(tilesFilterDIR, tileBaseName + "_filter.kea")
            filterImgsVals.append([tile, filterTile, filter_inst])

        with Pool(n_cores) as p:
            p.map(_perform_filtering_func, filterImgsVals)

        imgFilterTiles = glob.glob(os.path.join(tilesFilterDIR, "*_filter.kea"))

        numOutBands = rsgislib.imageutils.get_img_band_count(input_img)

        imageutils.create_copy_img(
            input_img, output_img, numOutBands, 0, gdalformat, datatype
        )

        imageutils.include_imgs_with_overlap(
            output_img, imgFilterTiles, int(filter_inst.getFilterHSize())
        )

        shutil.rmtree(tilesFilterDIR)

    shutil.rmtree(imgTilesDIR)
    if createdTmp:
        shutil.rmtree(tmp_dir)



#! /usr/bin/env python

############################################################################
# Copyright (c) 2019 Dr. Peter Bunting, Aberystwyth University
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
# 
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.
#
#
# Purpose:  A command to apply a command (e.g., gdal_translate) using subprocess 
#           to all files in an input directory while replicating the file structure 
#           of the input directory in the output directory.
#
# Author: Pete Bunting
# Email: pfb@aber.ac.uk
# Date: 11/09/2019
# Version: 1.0
#
# History:
# Version 1.0 - Created.
#
#############################################################################

import sys
import argparse
import pathlib
import os
import jinja2
import subprocess
from multiprocessing import Pool

def run_cmd(execmd):
    try:
        subprocess.call(execmd, shell=True)
    except OSError as e:
        raise Exception('Could not excute command: {}'.format(execmd))



if __name__ == '__main__':
    """
    The command line user interface
    """
    
    if len(sys.argv) == 1:
        print('rsgisapplycmd.py [options]')
        print('help : rsgisapplycmd.py --help')
        print('  or : rsgisapplycmd.py -h')
        print('')
        print('Example: rsgisapplycmd.py --indir ./KEA --outdir ./GTIFF --inext kea --outext tif --ncores 10 \\')
        print('                          --cmd "gdal_translate -of GTIFF {{ input }} {{ output }}"\n')
    
        print('Example: docker run -iv :/data petebunting/au-eoed rsgisapplycmd.py --indir /data/KEA \\')
        print('         --outdir /data/GTIFF --inext kea --outext tif --ncores 20 \\')
        print('         --cmd "gdal_translate -of GTIFF -co  TILED=YES -co COMPRESS=LZW -co COPY_SRC_OVERVIEWS=YES {{ input }} {{ output }}"\n')
    
    else:
        parser = argparse.ArgumentParser(prog='rsgisapplycmd.py', description='''Applies a command (e.g., gdal_translate) using 
                                                                                 subprocess to all files in an input directory
                                                                                 while replicating the file structure of the 
                                                                                 input directory in the output directory.''')
    
        parser.add_argument("--indir", type=str, required=True, help='''Input directory.''')
        parser.add_argument("--outdir", type=str, required=True, help='''Ouput directory.''')
        parser.add_argument("--inext", type=str, required=True, help='''Input file extension of the files to be processed (e.g., kea).''')
        parser.add_argument("--outext", type=str, required=True, help='''Output file extension for the output files (e.g., tif).''')
        parser.add_argument("--nameapp", type=str, required=False, default='', help='''Provide a string to be appended on the file 
                                                                                       name after the existing name and before the extension.''')
        parser.add_argument("--cmd", type=str, required=True, help='''The command to be excuted with {{ input }} and {{ output }} 
                                                                      specifying the where the input and output files should be.
                                                                      e.g., "gdal_translate -of GTIFF {{ input }} {{ output }}"''')
        parser.add_argument("--ncores", type=int, required=False, default=1, help='''Number of cores to used for processing (default = 1).''')
        parser.add_argument("--printcmds", action='store_true', default=False, help='''Prints the commands to be executed to the terminal rather than running.''')
        
    
        # Call the parser to parse the arguments.
        args = parser.parse_args()
    
        search_dir = args.indir
        search_dir_path = pathlib.Path(search_dir)
        output_dir = args.outdir
        output_dir_path = pathlib.Path(output_dir)
        
        in_ext = args.inext
        out_ext = args.outext
        out_app_name = args.nameapp
        
        cmds = []
        for root, dirs, files in os.walk(search_dir):
            for c_dir in dirs:
                c_path = pathlib.Path(os.path.join(root, c_dir))
                rel_to_root_path = c_path.relative_to(search_dir_path)
                out_dir_path = output_dir_path.joinpath(rel_to_root_path)
                if not os.path.exists(out_dir_path):
                   os.makedirs(out_dir_path)
            
            c_path = pathlib.Path(root)
            rel_to_root_path = c_path.relative_to(search_dir_path)
            out_dir_path = output_dir_path.joinpath(rel_to_root_path)
            for c_file in files:
                c_file_wpath = os.path.join(root, c_file)
                
                c_file_noext, c_ext = os.path.splitext(c_file)
                c_ext = c_ext[1:]
                if c_ext == in_ext:
                    o_file = '{0}{1}.{2}'.format(c_file_noext, out_app_name, out_ext)
                    o_file_wpath = os.path.join(out_dir_path, o_file)
                    
                    template = jinja2.Template(args.cmd)            
                    popd_cmd = template.render(input=c_file_wpath, output=o_file_wpath)
                    cmds.append(popd_cmd)
        
        if args.printcmds:
            for cmd in cmds:
                print(cmd)
        else:
            mpool = Pool(args.ncores)
            mpool.map(run_cmd, cmds)
                





#!/usr/bin/env python

#############################################
# rsgisfilehash.py
#
#  Copyright 2020 RSGISLib.
#
#  RSGISLib: 'The remote sensing and GIS Software Library'
#
#  RSGISLib is free software: you can redistribute it and/or modify
#  it under the terms of the GNU General Public License as published by
#  the Free Software Foundation, either version 3 of the License, or
#  (at your option) any later version.
#
#  RSGISLib is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#
#  You should have received a copy of the GNU General Public License
#  along with RSGISLib.  If not, see <http://www.gnu.org/licenses/>.
#
# Purpose: A script to find a hash string for an input file.
#
# Author: Pete Bunting
# Email: petebunting@mac.com
# Date: 08/04/2020
# Version: 1.0
#
##############################################

import os
import rsgislib.tools.filetools
import argparse

# Get input parameters
parser = argparse.ArgumentParser()
parser.add_argument("-i", "--input", type=str, required=True, help="Input file for which the hash will be created for.")
parser.add_argument("-b", '--blocksize', required=False, type=int, default=4096,
                    help="Block size used to calculate the file hash, default is 4096 bytes")
parser.add_argument("-o", '--outhash', required=False,
                    choices=['SHA1', 'SHA224', 'SHA256', 'SHA384', 'SHA512', 'MD5', 'Blake2B', 'Blake2S', 'SHA3_224',
                             'SHA3_256', 'SHA3_384', 'SHA3_512'], default='SHA256',
                    help="The hash to calculate; default: SHA256")

args = parser.parse_args()

input_file = args.input
if not os.path.exists(input_file):
    raise Exception("The input file does not exist.")

if args.outhash == 'SHA1':
    hash = rsgislib.tools.filetools.create_sha1_hash(input_file, args.blocksize)
elif args.outhash == 'SHA224':
    hash = rsgislib.tools.filetools.create_sha224_hash(input_file, args.blocksize)
elif args.outhash == 'SHA256':
    hash = rsgislib.tools.filetools.create_sha256_hash(input_file, args.blocksize)
elif args.outhash == 'SHA384':
    hash = rsgislib.tools.filetools.create_sha384_hash(input_file, args.blocksize)
elif args.outhash == 'SHA512':
    hash = rsgislib.tools.filetools.create_sha512_hash(input_file, args.blocksize)
elif args.outhash == 'MD5':
    hash = rsgislib.tools.filetools.create_md5_hash(input_file, args.blocksize)
elif args.outhash == 'Blake2B':
    hash = rsgislib.tools.filetools.create_blake2b_hash(input_file, args.blocksize)
elif args.outhash == 'Blake2S':
    hash = rsgislib.tools.filetools.create_blake2s_hash(input_file, args.blocksize)
elif args.outhash == 'SHA3_224':
    hash = rsgislib.tools.filetools.create_sha3_224_hash(input_file, args.blocksize)
elif args.outhash == 'SHA3_256':
    hash = rsgislib.tools.filetools.create_sha3_256_hash(input_file, args.blocksize)
elif args.outhash == 'SHA3_384':
    hash = rsgislib.tools.filetools.create_sha3_384_hash(input_file, args.blocksize)
elif args.outhash == 'SHA3_512':
    hash = rsgislib.tools.filetools.create_sha3_512_hash(input_file, args.blocksize)
else:
    raise Exception("The hash choice is unknown.")

print(hash)



import os
import pytest
from shutil import copy2

GEOPANDAS_NOT_AVAIL = False
try:
    import geopandas
except ImportError:
    GEOPANDAS_NOT_AVAIL = True

ALPHASHAPE_NOT_AVAIL = False
try:
    import alphashape
except ImportError:
    ALPHASHAPE_NOT_AVAIL = True

RTREE_NOT_AVAIL = False
try:
    import rtree
except ImportError:
    RTREE_NOT_AVAIL = True


DATA_DIR = os.path.join(os.path.dirname(os.path.abspath(__file__)), "data")
VECTORUTILS_DATA_DIR = os.path.join(DATA_DIR, "vectorutils")
VECTORGEOMS_DATA_DIR = os.path.join(DATA_DIR, "vectorgeoms")


def test_convert_polygon_to_polyline(tmp_path):
    import rsgislib.vectorgeoms

    vec_poly_file = os.path.join(DATA_DIR, "aber_osgb_multi_polys.geojson")
    vec_poly_lyr = "aber_osgb_multi_polys"

    vec_line_file = os.path.join(tmp_path, "out_vec.gpkg")
    vec_line_lyr = "out_vec"

    rsgislib.vectorgeoms.convert_polygon_to_polyline(
        vec_poly_file,
        vec_poly_lyr,
        vec_line_file,
        vec_line_lyr,
        out_format="GPKG",
        del_exist_vec=False,
    )

    assert os.path.exists(vec_line_file)


@pytest.mark.skipif(GEOPANDAS_NOT_AVAIL, reason="geopandas dependency not available")
def test_convert_polys_to_lines_gp(tmp_path):
    import rsgislib.vectorgeoms

    vec_poly_file = os.path.join(VECTORGEOMS_DATA_DIR, "welsh_land_area.gpkg")
    vec_poly_lyr = "welsh_land_area"

    vec_line_file = os.path.join(tmp_path, "out_vec.gpkg")
    vec_line_lyr = "out_vec"

    rsgislib.vectorgeoms.convert_polys_to_lines_gp(
        vec_poly_file,
        vec_poly_lyr,
        vec_line_file,
        vec_line_lyr,
        out_format="GPKG",
        del_exist_vec=False,
        exterior_lines=False,
    )

    assert os.path.exists(vec_line_file)


@pytest.mark.skipif(GEOPANDAS_NOT_AVAIL, reason="geopandas dependency not available")
def test_convert_polys_to_lines_gp_exterior(tmp_path):
    import rsgislib.vectorgeoms

    vec_poly_file = os.path.join(VECTORGEOMS_DATA_DIR, "welsh_land_area.gpkg")
    vec_poly_lyr = "welsh_land_area"

    vec_line_file = os.path.join(tmp_path, "out_vec.gpkg")
    vec_line_lyr = "out_vec"

    rsgislib.vectorgeoms.convert_polys_to_lines_gp(
        vec_poly_file,
        vec_poly_lyr,
        vec_line_file,
        vec_line_lyr,
        out_format="GPKG",
        del_exist_vec=False,
        exterior_lines=True,
    )

    assert os.path.exists(vec_line_file)


def test_get_pt_on_line():
    import rsgislib.vectorgeoms
    from osgeo import ogr

    pt1 = ogr.Geometry(ogr.wkbPoint)
    pt1.AddPoint(200, 500)
    pt2 = ogr.Geometry(ogr.wkbPoint)
    pt2.AddPoint(1200, 1500)

    new_pt_x, new_pt_y = rsgislib.vectorgeoms.get_pt_on_line(pt1, pt2, 250)

    assert (abs(new_pt_x - 376.7767) < 1) and (abs(new_pt_y - 676.7767) < 1)


def test_find_pt_to_side_right():
    import rsgislib.vectorgeoms
    from osgeo import ogr

    pt_start = ogr.Geometry(ogr.wkbPoint)
    pt_start.AddPoint(200, 500)

    pt = ogr.Geometry(ogr.wkbPoint)
    pt.AddPoint(376.7767, 676.7767)

    pt_end = ogr.Geometry(ogr.wkbPoint)
    pt_end.AddPoint(1200, 1500)

    new_pt_x, new_pt_y = rsgislib.vectorgeoms.find_pt_to_side(
        pt_start, pt, pt_end, 250, left_hand=False
    )

    assert (abs(new_pt_x - 553.5534) < 1) and (abs(new_pt_y - 500.000) < 1)


def test_find_pt_to_side_left():
    import rsgislib.vectorgeoms
    from osgeo import ogr

    pt_start = ogr.Geometry(ogr.wkbPoint)
    pt_start.AddPoint(200, 500)

    pt = ogr.Geometry(ogr.wkbPoint)
    pt.AddPoint(376.7767, 676.7767)

    pt_end = ogr.Geometry(ogr.wkbPoint)
    pt_end.AddPoint(1200, 1500)

    new_pt_x, new_pt_y = rsgislib.vectorgeoms.find_pt_to_side(
        pt_start, pt, pt_end, 250, left_hand=True
    )

    assert (abs(new_pt_x - 200.0000) < 1) and (abs(new_pt_y - 853.5534) < 1)


def test_create_orthg_lines_right(tmp_path):
    import rsgislib.vectorgeoms

    vec_line_file = os.path.join(VECTORGEOMS_DATA_DIR, "welsh_coastal_lines.geojson")
    vec_line_lyr = "welsh_coastal_lines"

    out_vec_file = os.path.join(tmp_path, "out_vec.gpkg")
    out_vec_lyr = "out_vec"

    rsgislib.vectorgeoms.create_orthg_lines(
        vec_line_file,
        vec_line_lyr,
        out_vec_file,
        out_vec_lyr,
        pt_step=1000,
        line_len=10000,
        left_hand=False,
        out_format="GPKG",
        del_exist_vec=False,
    )

    assert os.path.exists(out_vec_file)


def test_create_orthg_lines_left(tmp_path):
    import rsgislib.vectorgeoms

    vec_line_file = os.path.join(VECTORGEOMS_DATA_DIR, "welsh_coastal_lines.geojson")
    vec_line_lyr = "welsh_coastal_lines"

    out_vec_file = os.path.join(tmp_path, "out_vec.gpkg")
    out_vec_lyr = "out_vec"

    rsgislib.vectorgeoms.create_orthg_lines(
        vec_line_file,
        vec_line_lyr,
        out_vec_file,
        out_vec_lyr,
        pt_step=1000,
        line_len=10000,
        left_hand=True,
        out_format="GPKG",
        del_exist_vec=False,
    )

    assert os.path.exists(out_vec_file)


def test_closest_line_intersection(tmp_path):
    import rsgislib.vectorgeoms

    vec_line_file = os.path.join(
        VECTORGEOMS_DATA_DIR, "welsh_coastal_orthg_lines.geojson"
    )
    vec_line_lyr = "welsh_coastal_orthg_lines"

    vec_objs_file = os.path.join(VECTORGEOMS_DATA_DIR, "welsh_land_area.gpkg")
    vec_objs_lyr = "welsh_land_area"

    out_vec_file = os.path.join(tmp_path, "out_vec.gpkg")
    out_vec_lyr = "out_vec"

    rsgislib.vectorgeoms.closest_line_intersection(
        vec_line_file,
        vec_line_lyr,
        vec_objs_file,
        vec_objs_lyr,
        out_vec_file,
        out_vec_lyr,
        start_x_field="start_x",
        start_y_field="start_y",
        uid_field="uid",
        out_format="GPKG",
        del_exist_vec=False,
    )

    assert os.path.exists(out_vec_file)


def test_line_intersection_range(tmp_path):
    import rsgislib.vectorgeoms

    vec_line_file = os.path.join(
        VECTORGEOMS_DATA_DIR, "welsh_coastal_orthg_lines.geojson"
    )
    vec_line_lyr = "welsh_coastal_orthg_lines"

    vec_objs_file = os.path.join(VECTORGEOMS_DATA_DIR, "welsh_land_area.gpkg")
    vec_objs_lyr = "welsh_land_area"

    out_vec_file = os.path.join(tmp_path, "out_vec.gpkg")
    out_vec_lyr = "out_vec"

    rsgislib.vectorgeoms.line_intersection_range(
        vec_line_file,
        vec_line_lyr,
        vec_objs_file,
        vec_objs_lyr,
        out_vec_file,
        out_vec_lyr,
        start_x_field="start_x",
        start_y_field="start_y",
        uid_field="uid",
        out_format="GPKG",
        del_exist_vec=False,
    )

    assert os.path.exists(out_vec_file)


def test_scnd_line_intersection_range(tmp_path):
    import rsgislib.vectorgeoms

    vec_line_file = os.path.join(
        VECTORGEOMS_DATA_DIR, "welsh_coastal_orthg_lines.geojson"
    )
    vec_line_lyr = "welsh_coastal_orthg_lines"

    vec_objs_file = os.path.join(VECTORGEOMS_DATA_DIR, "welsh_land_area.gpkg")
    vec_objs_lyr = "welsh_land_area"

    out_vec_file = os.path.join(tmp_path, "out_vec.gpkg")
    out_vec_lyr = "out_vec"

    rsgislib.vectorgeoms.scnd_line_intersection_range(
        vec_line_file,
        vec_line_lyr,
        vec_objs_file,
        vec_objs_lyr,
        out_vec_file,
        out_vec_lyr,
        start_x_field="start_x",
        start_y_field="start_y",
        uid_field="uid",
        out_format="GPKG",
        del_exist_vec=False,
    )

    assert os.path.exists(out_vec_file)


def test_calc_poly_centroids(tmp_path):
    import rsgislib.vectorgeoms

    vec_file = os.path.join(DATA_DIR, "aber_osgb_multi_polys.geojson")
    vec_lyr = "aber_osgb_multi_polys"

    out_vec_file = os.path.join(tmp_path, "out_vec.gpkg")
    out_vec_lyr = "out_vec"

    rsgislib.vectorgeoms.calc_poly_centroids(
        vec_file, vec_lyr, "GPKG", out_vec_file, out_vec_lyr
    )

    assert os.path.exists(out_vec_file)


@pytest.mark.skipif(GEOPANDAS_NOT_AVAIL, reason="geopandas dependency not available")
def test_vec_lyr_intersection_gp(tmp_path):
    import rsgislib.vectorgeoms

    vec_file = os.path.join(DATA_DIR, "aber_osgb_multi_polys.geojson")
    vec_lyr = "aber_osgb_multi_polys"

    vec_over_file = os.path.join(VECTORGEOMS_DATA_DIR, "welsh_land_area.gpkg")
    vec_over_lyr = "welsh_land_area"

    out_vec_file = os.path.join(tmp_path, "out_vec.gpkg")
    out_vec_lyr = "out_vec"

    rsgislib.vectorgeoms.vec_lyr_intersection_gp(
        vec_file,
        vec_lyr,
        vec_over_file,
        vec_over_lyr,
        out_vec_file,
        out_vec_lyr,
        out_format="GPKG",
        del_exist_vec=False,
    )

    assert os.path.exists(out_vec_file)


@pytest.mark.skipif(GEOPANDAS_NOT_AVAIL, reason="geopandas dependency not available")
def test_vec_lyr_difference_gp(tmp_path):
    import rsgislib.vectorgeoms

    vec_file = os.path.join(DATA_DIR, "aber_osgb_multi_polys.geojson")
    vec_lyr = "aber_osgb_multi_polys"

    vec_over_file = os.path.join(VECTORGEOMS_DATA_DIR, "welsh_land_area.gpkg")
    vec_over_lyr = "welsh_land_area"

    out_vec_file = os.path.join(tmp_path, "out_vec.gpkg")
    out_vec_lyr = "out_vec"

    rsgislib.vectorgeoms.vec_lyr_difference_gp(
        vec_file,
        vec_lyr,
        vec_over_file,
        vec_over_lyr,
        out_vec_file,
        out_vec_lyr,
        out_format="GPKG",
        del_exist_vec=False,
    )

    assert os.path.exists(out_vec_file)


@pytest.mark.skipif(GEOPANDAS_NOT_AVAIL, reason="geopandas dependency not available")
def test_vec_lyr_sym_difference_gp(tmp_path):
    import rsgislib.vectorgeoms

    vec_file = os.path.join(DATA_DIR, "aber_osgb_multi_polys.geojson")
    vec_lyr = "aber_osgb_multi_polys"

    vec_over_file = os.path.join(VECTORGEOMS_DATA_DIR, "welsh_land_area.gpkg")
    vec_over_lyr = "welsh_land_area"

    out_vec_file = os.path.join(tmp_path, "out_vec.gpkg")
    out_vec_lyr = "out_vec"

    rsgislib.vectorgeoms.vec_lyr_sym_difference_gp(
        vec_file,
        vec_lyr,
        vec_over_file,
        vec_over_lyr,
        out_vec_file,
        out_vec_lyr,
        out_format="GPKG",
        del_exist_vec=False,
    )

    assert os.path.exists(out_vec_file)


@pytest.mark.skipif(GEOPANDAS_NOT_AVAIL, reason="geopandas dependency not available")
def test_vec_lyr_identity_gp(tmp_path):
    import rsgislib.vectorgeoms

    vec_file = os.path.join(DATA_DIR, "aber_osgb_multi_polys.geojson")
    vec_lyr = "aber_osgb_multi_polys"

    vec_over_file = os.path.join(VECTORGEOMS_DATA_DIR, "welsh_land_area.gpkg")
    vec_over_lyr = "welsh_land_area"

    out_vec_file = os.path.join(tmp_path, "out_vec.gpkg")
    out_vec_lyr = "out_vec"

    rsgislib.vectorgeoms.vec_lyr_identity_gp(
        vec_file,
        vec_lyr,
        vec_over_file,
        vec_over_lyr,
        out_vec_file,
        out_vec_lyr,
        out_format="GPKG",
        del_exist_vec=False,
    )

    assert os.path.exists(out_vec_file)


@pytest.mark.skipif(GEOPANDAS_NOT_AVAIL, reason="geopandas dependency not available")
def test_vec_lyr_union_gp(tmp_path):
    import rsgislib.vectorgeoms

    vec_file = os.path.join(DATA_DIR, "aber_osgb_multi_polys.geojson")
    vec_lyr = "aber_osgb_multi_polys"

    vec_over_file = os.path.join(VECTORGEOMS_DATA_DIR, "welsh_land_area.gpkg")
    vec_over_lyr = "welsh_land_area"

    out_vec_file = os.path.join(tmp_path, "out_vec.gpkg")
    out_vec_lyr = "out_vec"

    rsgislib.vectorgeoms.vec_lyr_union_gp(
        vec_file,
        vec_lyr,
        vec_over_file,
        vec_over_lyr,
        out_vec_file,
        out_vec_lyr,
        out_format="GPKG",
        del_exist_vec=False,
    )

    assert os.path.exists(out_vec_file)


def test_get_vec_lyr_as_pts():
    import rsgislib.vectorgeoms

    vec_file = os.path.join(DATA_DIR, "aber_osgb_single_poly.geojson")
    vec_lyr = "aber_osgb_single_poly"

    pts = rsgislib.vectorgeoms.get_vec_lyr_as_pts(vec_file, vec_lyr)

    assert len(pts) == 41


@pytest.mark.skipif(ALPHASHAPE_NOT_AVAIL, reason="alphashape dependency not available")
def test_create_alpha_shape(tmp_path):
    import rsgislib.vectorgeoms

    vec_file = os.path.join(DATA_DIR, "aber_osgb_single_poly.geojson")
    vec_lyr = "aber_osgb_single_poly"

    out_vec_file = os.path.join(tmp_path, "out_vec.gpkg")
    out_vec_lyr = "out_vec"

    rsgislib.vectorgeoms.create_alpha_shape(
        vec_file,
        vec_lyr,
        out_vec_file,
        out_vec_lyr,
        out_format="GPKG",
        alpha_val=2,
        alpha_vals=None,
        max_iter=100,
        del_exist_vec=False,
    )

    assert os.path.exists(out_vec_file)


@pytest.mark.skipif(GEOPANDAS_NOT_AVAIL, reason="geopandas dependency not available")
def test_explode_vec_lyr(tmp_path):
    import rsgislib.vectorgeoms

    vec_file = os.path.join(VECTORGEOMS_DATA_DIR, "welsh_land_area.gpkg")
    vec_lyr = "welsh_land_area"

    out_vec_file = os.path.join(tmp_path, "out_vec.gpkg")
    out_vec_lyr = "out_vec"

    rsgislib.vectorgeoms.explode_vec_lyr(
        vec_file, vec_lyr, out_vec_file, out_vec_lyr, out_format="GPKG"
    )

    assert os.path.exists(out_vec_file)


@pytest.mark.skipif(GEOPANDAS_NOT_AVAIL, reason="geopandas dependency not available")
def test_explode_vec_files(tmp_path):
    import rsgislib.vectorgeoms
    import glob

    vec_file = os.path.join(VECTORGEOMS_DATA_DIR, "welsh_land_area.gpkg")

    rsgislib.vectorgeoms.explode_vec_files([vec_file], output_dir=tmp_path)

    assert len(glob.glob(os.path.join(tmp_path, "*.gpkg"))) == 1


def test_convert_multi_geoms_to_single(tmp_path):
    import rsgislib.vectorgeoms

    vec_file = os.path.join(VECTORGEOMS_DATA_DIR, "welsh_land_area.gpkg")
    vec_lyr = "welsh_land_area"

    out_vec_file = os.path.join(tmp_path, "out_vec.gpkg")
    out_vec_lyr = "out_vec"

    rsgislib.vectorgeoms.convert_multi_geoms_to_single(
        vec_file, vec_lyr, "GPKG", out_vec_file, out_vec_lyr
    )

    assert os.path.exists(out_vec_file)


def test_simplify_geometries(tmp_path):
    import rsgislib.vectorgeoms

    vec_file = os.path.join(DATA_DIR, "aber_osgb_single_poly.geojson")
    vec_lyr = "aber_osgb_single_poly"

    out_vec_file = os.path.join(tmp_path, "out_vec.gpkg")
    out_vec_lyr = "out_vec"

    rsgislib.vectorgeoms.simplify_geometries(
        vec_file, vec_lyr, 20, "GPKG", out_vec_file, out_vec_lyr, del_exist_vec=False
    )

    assert os.path.exists(out_vec_file)


def test_delete_polygon_holes(tmp_path):
    import rsgislib.vectorgeoms

    vec_file = os.path.join(DATA_DIR, "aber_osgb_single_poly_hole.geojson")
    vec_lyr = "aber_osgb_single_poly_hole"

    out_vec_file = os.path.join(tmp_path, "out_vec.gpkg")
    out_vec_lyr = "out_vec"

    rsgislib.vectorgeoms.delete_polygon_holes(
        vec_file,
        vec_lyr,
        "GPKG",
        out_vec_file,
        out_vec_lyr,
        area_thres=None,
        del_exist_vec=False,
    )

    assert os.path.exists(out_vec_file)


def test_get_poly_hole_area():
    import rsgislib.vectorgeoms

    vec_file = os.path.join(DATA_DIR, "aber_osgb_single_poly_hole.geojson")
    vec_lyr = "aber_osgb_single_poly_hole"

    areas = rsgislib.vectorgeoms.get_poly_hole_area(vec_file, vec_lyr)

    assert (len(areas) == 1) and (abs(areas[0] - 524701) < 1)


def test_remove_polygon_area(tmp_path):
    import rsgislib.vectorgeoms

    vec_file = os.path.join(DATA_DIR, "aber_osgb_multi_polys.geojson")
    vec_lyr = "aber_osgb_multi_polys"

    out_vec_file = os.path.join(tmp_path, "out_vec.gpkg")
    out_vec_lyr = "out_vec"

    rsgislib.vectorgeoms.remove_polygon_area(
        vec_file,
        vec_lyr,
        "GPKG",
        out_vec_file,
        out_vec_lyr,
        area_thres=10000,
        del_exist_vec=False,
    )

    assert os.path.exists(out_vec_file)


def test_vec_lyr_intersection(tmp_path):
    import rsgislib.vectorgeoms

    vec_file = os.path.join(DATA_DIR, "aber_osgb_multi_polys.geojson")
    vec_lyr = "aber_osgb_multi_polys"

    vec_over_file = os.path.join(VECTORGEOMS_DATA_DIR, "welsh_land_area.gpkg")
    vec_over_lyr = "welsh_land_area"

    out_vec_file = os.path.join(tmp_path, "out_vec.gpkg")
    out_vec_lyr = "out_vec"

    rsgislib.vectorgeoms.vec_lyr_intersection(
        vec_file,
        vec_lyr,
        vec_over_file,
        vec_over_lyr,
        out_vec_file,
        out_vec_lyr,
        out_format="GPKG",
        del_exist_vec=False,
    )

    assert os.path.exists(out_vec_file)


def test_vec_lyr_difference(tmp_path):
    import rsgislib.vectorgeoms

    vec_file = os.path.join(DATA_DIR, "aber_osgb_multi_polys.geojson")
    vec_lyr = "aber_osgb_multi_polys"

    vec_over_file = os.path.join(VECTORGEOMS_DATA_DIR, "welsh_land_area.gpkg")
    vec_over_lyr = "welsh_land_area"

    out_vec_file = os.path.join(tmp_path, "out_vec.gpkg")
    out_vec_lyr = "out_vec"

    rsgislib.vectorgeoms.vec_lyr_difference(
        vec_file,
        vec_lyr,
        vec_over_file,
        vec_over_lyr,
        out_vec_file,
        out_vec_lyr,
        out_format="GPKG",
        del_exist_vec=False,
    )

    assert os.path.exists(out_vec_file)


def test_vec_lyr_difference_sym(tmp_path):
    import rsgislib.vectorgeoms

    vec_file = os.path.join(DATA_DIR, "aber_osgb_multi_polys.geojson")
    vec_lyr = "aber_osgb_multi_polys"

    vec_over_file = os.path.join(VECTORGEOMS_DATA_DIR, "welsh_land_area.gpkg")
    vec_over_lyr = "welsh_land_area"

    out_vec_file = os.path.join(tmp_path, "out_vec.gpkg")
    out_vec_lyr = "out_vec"

    rsgislib.vectorgeoms.vec_lyr_difference(
        vec_file,
        vec_lyr,
        vec_over_file,
        vec_over_lyr,
        out_vec_file,
        out_vec_lyr,
        out_format="GPKG",
        symmetric=True,
        del_exist_vec=False,
    )

    assert os.path.exists(out_vec_file)


@pytest.mark.skipif(GEOPANDAS_NOT_AVAIL, reason="geopandas dependency not available")
def test_clip_vec_lyr(tmp_path):
    import rsgislib.vectorgeoms

    vec_file = os.path.join(VECTORGEOMS_DATA_DIR, "welsh_land_area.gpkg")
    vec_lyr = "welsh_land_area"

    vec_roi_file = os.path.join(DATA_DIR, "aber_osgb_multi_polys.geojson")
    vec_roi_lyr = "aber_osgb_multi_polys"

    out_vec_file = os.path.join(tmp_path, "out_vec.gpkg")
    out_vec_lyr = "out_vec"

    rsgislib.vectorgeoms.clip_vec_lyr(
        vec_file,
        vec_lyr,
        vec_roi_file,
        vec_roi_lyr,
        out_vec_file,
        out_vec_lyr,
        out_format="GPKG",
    )

    assert os.path.exists(out_vec_file)


def test_get_geom_pts():
    import rsgislib.vectorgeoms
    from osgeo import ogr

    # Create ring
    ring = ogr.Geometry(ogr.wkbLinearRing)
    ring.AddPoint(1179091.1646903288, 712782.8838459781)
    ring.AddPoint(1161053.0218226474, 667456.2684348812)
    ring.AddPoint(1214704.933941905, 641092.8288590391)
    ring.AddPoint(1228580.428455506, 682719.3123998424)
    ring.AddPoint(1218405.0658121984, 721108.1805541387)
    ring.AddPoint(1179091.1646903288, 712782.8838459781)

    # Create polygon
    poly = ogr.Geometry(ogr.wkbPolygon)
    poly.AddGeometry(ring)

    pts = rsgislib.vectorgeoms.get_geom_pts(poly)

    assert len(pts) == 6


def test_vec_intersects_vec():
    import rsgislib.vectorgeoms

    vec_file = os.path.join(VECTORGEOMS_DATA_DIR, "welsh_land_area.gpkg")
    vec_lyr = "welsh_land_area"

    vec_roi_file = os.path.join(DATA_DIR, "aber_osgb_multi_polys.geojson")
    vec_roi_lyr = "aber_osgb_multi_polys"

    out_bool = rsgislib.vectorgeoms.vec_intersects_vec(
        vec_file, vec_lyr, vec_roi_file, vec_roi_lyr
    )

    assert out_bool


def test_vec_overlaps_vec():
    import rsgislib.vectorgeoms

    vec_file = os.path.join(VECTORGEOMS_DATA_DIR, "welsh_land_area.gpkg")
    vec_lyr = "welsh_land_area"

    vec_roi_file = os.path.join(DATA_DIR, "aber_osgb_multi_polys.geojson")
    vec_roi_lyr = "aber_osgb_multi_polys"

    out_bool = rsgislib.vectorgeoms.vec_overlaps_vec(
        vec_file, vec_lyr, vec_roi_file, vec_roi_lyr
    )

    assert out_bool


def test_vec_within_vec():
    import rsgislib.vectorgeoms

    vec_file = os.path.join(VECTORGEOMS_DATA_DIR, "welsh_land_area.gpkg")
    vec_lyr = "welsh_land_area"

    vec_roi_file = os.path.join(DATA_DIR, "aber_osgb_multi_polys.geojson")
    vec_roi_lyr = "aber_osgb_multi_polys"

    out_bool = rsgislib.vectorgeoms.vec_within_vec(
        vec_file, vec_lyr, vec_roi_file, vec_roi_lyr
    )

    assert not out_bool


def test_vec_contains_vec():
    import rsgislib.vectorgeoms

    vec_file = os.path.join(VECTORGEOMS_DATA_DIR, "welsh_land_area.gpkg")
    vec_lyr = "welsh_land_area"

    vec_roi_file = os.path.join(DATA_DIR, "aber_osgb_multi_polys.geojson")
    vec_roi_lyr = "aber_osgb_multi_polys"

    out_bool = rsgislib.vectorgeoms.vec_contains_vec(
        vec_file, vec_lyr, vec_roi_file, vec_roi_lyr
    )

    assert not out_bool


def test_vec_touches_vec():
    import rsgislib.vectorgeoms

    vec_file = os.path.join(VECTORGEOMS_DATA_DIR, "welsh_land_area.gpkg")
    vec_lyr = "welsh_land_area"

    vec_roi_file = os.path.join(DATA_DIR, "aber_osgb_multi_polys.geojson")
    vec_roi_lyr = "aber_osgb_multi_polys"

    out_bool = rsgislib.vectorgeoms.vec_touches_vec(
        vec_file, vec_lyr, vec_roi_file, vec_roi_lyr
    )

    assert not out_bool


def test_vec_crosses_vec():
    import rsgislib.vectorgeoms

    vec_file = os.path.join(VECTORGEOMS_DATA_DIR, "welsh_land_area.gpkg")
    vec_lyr = "welsh_land_area"

    vec_roi_file = os.path.join(DATA_DIR, "aber_osgb_multi_polys.geojson")
    vec_roi_lyr = "aber_osgb_multi_polys"

    out_bool = rsgislib.vectorgeoms.vec_crosses_vec(
        vec_file, vec_lyr, vec_roi_file, vec_roi_lyr
    )

    assert not out_bool


@pytest.mark.skipif(GEOPANDAS_NOT_AVAIL, reason="geopandas dependency not available")
def test_get_geoms_as_bboxs():
    import rsgislib.vectorgeoms

    vec_file = os.path.join(DATA_DIR, "aber_osgb_multi_polys.geojson")
    vec_lyr = "aber_osgb_multi_polys"

    bboxes = rsgislib.vectorgeoms.get_geoms_as_bboxs(vec_file, vec_lyr)

    assert len(bboxes) == 4


def test_bbox_intersects_vec_lyr():
    import rsgislib.vectorgeoms

    vec_file = os.path.join(VECTORGEOMS_DATA_DIR, "welsh_land_area.gpkg")
    vec_lyr = "welsh_land_area"

    bbox = [257774, 261752, 279153, 283287]

    out_bool = rsgislib.vectorgeoms.bbox_intersects_vec_lyr(vec_file, vec_lyr, bbox)

    assert out_bool


@pytest.mark.skipif(GEOPANDAS_NOT_AVAIL, reason="geopandas dependency not available")
def test_shiftxy_vec_lyr(tmp_path):
    import rsgislib.vectorgeoms

    vec_file = os.path.join(DATA_DIR, "aber_osgb_multi_polys.geojson")
    vec_lyr = "aber_osgb_multi_polys"

    out_vec_file = os.path.join(tmp_path, "out_vec.gpkg")
    out_vec_lyr = "out_vec"

    rsgislib.vectorgeoms.shiftxy_vec_lyr(
        vec_file, vec_lyr, 200, 300, out_vec_file, out_vec_lyr, out_format="GPKG"
    )

    assert os.path.exists(out_vec_file)


@pytest.mark.skipif(RTREE_NOT_AVAIL, reason="tree dependency not available")
def test_create_rtree_index():
    import rsgislib.vectorgeoms

    vec_file = os.path.join(DATA_DIR, "aber_osgb_multi_polys.geojson")
    vec_lyr = "aber_osgb_multi_polys"

    idx_obj, geom_lst = rsgislib.vectorgeoms.create_rtree_index(vec_file, vec_lyr)

    assert len(geom_lst) == 4


@pytest.mark.skipif(RTREE_NOT_AVAIL, reason="tree dependency not available")
def test_bbox_intersects_index():
    import rsgislib.vectorgeoms

    vec_file = os.path.join(DATA_DIR, "aber_osgb_multi_polys.geojson")
    vec_lyr = "aber_osgb_multi_polys"

    rt_idx, geom_lst = rsgislib.vectorgeoms.create_rtree_index(vec_file, vec_lyr)

    bbox = [257774, 261752, 279153, 283287]

    out_bool = rsgislib.vectorgeoms.bbox_intersects_index(rt_idx, geom_lst, bbox)

    assert out_bool
